import json
import os
import sys
import tarfile

from metaflow.metaflow_environment import MetaflowEnvironment
from metaflow.exception import MetaflowException
from metaflow.mflog import BASH_SAVE_LOGS

from .conda import Conda
from . import get_conda_manifest_path, CONDA_MAGIC_FILE


class CondaEnvironment(MetaflowEnvironment):
    TYPE = 'conda'
    _filecache = None

    def __init__(self, flow):
        self.flow = flow
        self.local_root = None
        # A conda environment sits on top of whatever default environment
        # the user has so we get that environment to be able to forward
        # any calls we don't handle specifically to that one.
        from ...plugins import ENVIRONMENTS
        from metaflow.metaflow_config import DEFAULT_ENVIRONMENT
        self.base_env = [e for e in ENVIRONMENTS + [MetaflowEnvironment]
            if e.TYPE == DEFAULT_ENVIRONMENT][0](self.flow)

    def init_environment(self, echo):
        # Print a message for now
        echo("Bootstrapping conda environment..." +
            "(this could take a few minutes)")
        self.base_env.init_environment(echo)

    def validate_environment(self, echo):
        return self.base_env.validate_environment(echo)

    def decospecs(self):
        # Apply conda decorator to all steps
        return ('conda', )

    def _get_conda_decorator(self, step_name):
        step = next(step for step in self.flow if step.name == step_name)
        decorator = next(deco for deco in step.decorators if deco.name == 'conda')
        # Guaranteed to have a conda decorator because of self.decospecs()
        return decorator

    def _get_env_id(self, step_name):
        conda_decorator = self._get_conda_decorator(step_name)
        if conda_decorator.is_enabled():
            return conda_decorator._env_id()
        return None

    def _get_executable(self, step_name):
        env_id = self._get_env_id(step_name)
        if env_id is not None:
            return (os.path.join(env_id, "bin/python -s"))
        return None

    def set_local_root(self, ds_root):
        self.local_root = ds_root

    def bootstrap_commands(self, step_name):
        # Bootstrap conda and execution environment for step
        env_id = self._get_env_id(step_name)
        if env_id is not None:
            return [
                    "echo \'Bootstrapping environment...\'",
                    "python -m metaflow.plugins.conda.batch_bootstrap \"%s\" %s" % \
                        (self.flow.name, env_id),
                    "echo \'Environment bootstrapped.\'",
                ]
        return []

    def add_to_package(self):
        files = self.base_env.add_to_package()
        # Add conda manifest file to job package at the top level.
        path = get_conda_manifest_path(self.local_root, self.flow.name)
        if os.path.exists(path):
            files.append((path, os.path.basename(path)))
        return files

    def pylint_config(self):
        config = self.base_env.pylint_config()
        # Disable (import-error) in pylint
        config.append('--disable=F0401')
        return config

    def executable(self, step_name):
        # Get relevant python interpreter for step
        executable = self._get_executable(step_name)
        if executable is not None:
            return executable
        return self.base_env.executable(step_name)

    @classmethod
    def get_client_info(cls, flow_name, metadata):
        if cls._filecache is None:
            from metaflow.client.filecache import FileCache
            cls._filecache = FileCache()
        info = metadata.get('code-package')
        env_id = metadata.get('conda_env_id')
        if info is None or env_id is None:
            return {'type': 'conda'}
        info = json.loads(info)
        with cls._filecache.get_data(info['ds_type'], flow_name, info['sha']) as f:
            tar = tarfile.TarFile(fileobj=f)
            conda_file = tar.extractfile(CONDA_MAGIC_FILE)
            if conda_file is None:
                return {'type': 'conda'}
            info = json.loads(conda_file.read().decode('utf-8'))
        new_info = {
            'type': 'conda',
            'explicit': info[env_id]['explicit'],
            'deps': info[env_id]['deps']}
        return new_info

    def get_package_commands(self, code_package_url):
        return self.base_env.get_package_commands(code_package_url)

    def get_environment_info(self):
        return self.base_env.get_environment_info()



"""
Local storage

Store data under .metaflow/ in the cwd
"""
import os
import json
import gzip
from tempfile import NamedTemporaryFile

from metaflow.util import Path
from metaflow.metaflow_config import DATASTORE_LOCAL_DIR, DATASTORE_SYSROOT_LOCAL
from .datastore import MetaflowDataStore, DataException, only_if_not_done
from ..metadata import MetaDatum


class LocalDataStore(MetaflowDataStore):
    TYPE = 'local'

    METADATA_DIR = '_meta'

    def _makedirs(self, path):
        try:
            os.makedirs(path)
        except OSError as x:
            if x.errno == 17:
                return
            else:
                raise

    def object_path(self, sha):
        root = os.path.join(self.data_root, sha[:2])
        return os.path.join(root, sha)

    @classmethod
    def get_datastore_root_from_config(cls, echo, create_on_absent=True):
        # Compute path for DATASTORE_SYSROOT_LOCAL
        result = DATASTORE_SYSROOT_LOCAL
        if result is None:
            try:
                # Python2
                current_path = os.getcwdu()
            except: # noqa E722
                current_path = os.getcwd()
            check_dir = os.path.join(current_path, DATASTORE_LOCAL_DIR)
            check_dir = os.path.realpath(check_dir)
            orig_path = check_dir
            top_level_reached = False
            while not os.path.isdir(check_dir):
                new_path = os.path.dirname(current_path)
                if new_path == current_path:
                    top_level_reached = True
                    break  # We are no longer making upward progress
                current_path = new_path
                check_dir = os.path.join(current_path, DATASTORE_LOCAL_DIR)
            if top_level_reached:
                if create_on_absent:
                    # Could not find any directory to use so create a new one
                    echo('Creating local datastore in current directory (%s)' % orig_path,
                         fg='magenta', bold=True)
                    os.mkdir(orig_path)
                    result = orig_path
                else:
                    return None
            else:
                result = check_dir
        else:
            result = os.path.join(result, DATASTORE_LOCAL_DIR)
        return result

    @classmethod
    def get_latest_tasks(cls,
                         flow_name,
                         run_id=None,
                         steps=None,
                         pathspecs=None):
        run_prefix = cls.make_path(flow_name, run_id)
        data_blobs = []

        if os.path.exists(run_prefix):
            if steps is None:
                steps = [s for s in os.listdir(run_prefix) if s != cls.METADATA_DIR]
            if pathspecs is None:
                task_prefixes = []
                for step in steps:
                    step_prefix = cls.make_path(flow_name, run_id, step)
                    for task in os.listdir(step_prefix):
                        if task == cls.METADATA_DIR:
                            continue
                        task_prefixes.append(
                            cls.make_path(flow_name, run_id, step, task))
            else:
                task_prefixes = [cls.make_path(flow_name, pathspec)
                                 for pathspec in pathspecs]
            for task_prefix in task_prefixes:
                step, task = task_prefix.split('/')[-2:]
                # Sort the file listing to iterate in increasing order of
                # attempts.
                latest_data_path = None
                latest_attempt = None
                latest_done_attempt = None
                for fname in sorted(os.listdir(task_prefix)):
                    if cls.is_done_filename(fname):
                        _, attempt = cls.parse_filename(fname)
                        latest_done_attempt = attempt
                        # Read the corresponding metadata file.
                        meta_fname = \
                            cls.get_metadata_filename_for_attempt(attempt)
                        latest_data_path = os.path.join(task_prefix, meta_fname)
                    elif cls.is_attempt_filename(fname):
                        _, attempt = cls.parse_filename(fname)
                        latest_attempt = attempt
                # Only read the metadata if the latest attempt is also done.
                if latest_done_attempt is not None and\
                    latest_done_attempt == latest_attempt:
                    with open(latest_data_path) as f:
                        data_blobs.append((step, task, attempt, f.read()))
            return data_blobs
        else:
            raise DataException("Couldn't find data at %s" % run_prefix)

    @classmethod
    def get_artifacts(cls, artifacts_to_prefetch):
        artifact_list = []
        for path in artifacts_to_prefetch:
            sha = path.split('/')[-1]
            artifact_list.append((sha,
                                  cls.decode_gzip_data(path)))
        return artifact_list

    @only_if_not_done
    def save_logs(self, logsource, stream_data):
        """
        Save log files for multiple streams, represented as
        as a list of (stream, bytes) or (stream, Path) tuples.
        """
        for stream, data in stream_data:
            if isinstance(data, Path):
                with open(str(data), 'rb') as f:
                    data = f.read()
            path = self.get_log_location(logsource, stream)
            with open(path + '.tmp', 'wb') as f:
                f.write(data)
            os.rename(path + '.tmp', path)

    def _read_file_or_empty(self, path):
        if os.path.exists(path):
            with open(path, 'rb') as f:
                return f.read()
        else:
            return b''

    def load_log_legacy(self, stream, attempt_override=None):
        """
        Load old-style, pre-mflog, log file represented as a bytes object.
        """
        f = self.filename_with_attempt_prefix('%s.log' % stream,
                attempt_override if attempt_override is not None 
                    else self.attempt)
        return self._read_file_or_empty(os.path.join(self.root, f))

    def load_logs(self, logsources, stream, attempt_override=None):
        paths = [self.get_log_location(source, stream, attempt_override) 
                    for source in logsources]
        return list(zip(logsources, map(self._read_file_or_empty, paths)))

    @only_if_not_done
    def save_metadata(self, name, metadata):
        """
        Save a task-specific metadata dictionary as JSON.
        """
        self._makedirs(self.root)
        filename = self.filename_with_attempt_prefix('%s.json' % name,
                                                     self.attempt)
        path = os.path.join(self.root, filename)
        with open(path + '.tmp', 'w') as f:
            json.dump(metadata, f)
        os.rename(path + '.tmp', path)

    def load_metadata(self, name):
        """
        Load a task-specific metadata dictionary as JSON.
        """
        filename = self.filename_with_attempt_prefix('%s.json' % name,
                                                     self.attempt)
        path = os.path.join(self.root, filename)
        with open(path) as f:
            return json.load(f)

    def has_metadata(self, name, with_attempt=True):
        attempt = self.attempt if with_attempt else None
        filename = self.filename_with_attempt_prefix('%s.json' % name, attempt)
        path = os.path.join(self.root, filename)
        return os.path.exists(path)

    @only_if_not_done
    def save_data(self, sha, transformable_object):
        """
        Save a content-addressed data blob if it doesn't exist already.
        """
        path = self.object_path(sha)
        if not os.path.exists(path):
            self._makedirs(os.path.dirname(path))
            # NOTE multiple tasks may try to save an object with the
            # same sha concurrently, hence we need to use a proper tmp
            # file
            with NamedTemporaryFile(dir=os.path.dirname(path),
                                    prefix='blobtmp.',
                                    delete=False) as tmp:
                # NOTE compresslevel makes a huge difference. The default
                # level of 9 can be impossibly slow.
                with gzip.GzipFile(fileobj=tmp,
                                   mode='wb',
                                   compresslevel=3) as f:
                    f.write(transformable_object.current())
            os.rename(tmp.name, path)
        return path

    def load_data(self, sha):
        """
        Load a content-addressed data blob.
        """
        with gzip.open(self.object_path(sha), 'rb') as f:
            return f.read()

    @only_if_not_done
    def done(self):
        """
        Write a marker indicating that datastore has finished writing to
        this path.
        """
        filename = self.get_done_filename_for_attempt(self.attempt)
        path = os.path.join(self.root, filename)
        self._makedirs(self.root)
        try:
            # this is for python2 compatibility.
            # Python3 has open(mode='x').
            fd = os.fdopen(os.open(path,
                                   os.O_EXCL | os.O_WRONLY | os.O_CREAT),
                                   'wb')
            fd.close()
        except OSError as x:
            if x.errno == 17:
                raise DataException('Path %s already exists. Try with a '
                                    'different --run-id.' % path)
            else:
                raise
        self.metadata.register_metadata(
            self.run_id, self.step_name, self.task_id,
            [MetaDatum(field='attempt-done', value=str(self.attempt), type='attempt-done', tags=[])])

        self._is_done_set = True

    def is_done(self):
        """
        A flag indicating whether this datastore directory was closed
        succesfully with done().
        """
        filename = self.get_done_filename_for_attempt(self.attempt)
        path = os.path.join(self.root, filename)
        return os.path.exists(path)



import json

from metaflow.util import to_unicode

"""
MetaflowDatastoreSet allows you to prefetch multiple (read) datastores into a
cache and lets you access them.
As a performance optimization it also lets you prefetch select data artifacts
leveraging a shared cache.
"""
class MetaflowDatastoreSet(object):
    def __init__(self,
                 ds_class,
                 flow_name,
                 run_id,
                 steps=None,
                 pathspecs=None,
                 metadata=None,
                 event_logger=None,
                 monitor=None,
                 prefetch_data_artifacts=None):
        data_blobs = ds_class.get_latest_tasks(flow_name,
                                               run_id,
                                               steps=steps,
                                               pathspecs=pathspecs)
        artifact_cache = {}
        datastores = [ds_class(flow_name,
                               run_id=run_id,
                               step_name=step_name,
                               task_id=task_id,
                               metadata=metadata,
                               attempt=attempt,
                               event_logger=event_logger,
                               monitor=monitor,
                               data_obj=json.loads(to_unicode(data_blob)),
                               artifact_cache=artifact_cache)
                      for step_name, task_id, attempt, data_blob in data_blobs]
        if prefetch_data_artifacts:
            artifacts_to_prefetch = set(
                [ds.artifact_path(artifact_name)
                 for ds in datastores
                 for artifact_name in prefetch_data_artifacts
                 if artifact_name in ds])

            # Update (and not re-assign) the artifact_cache since each datastore
            # created above has a reference to this object.
            artifact_cache.update(ds_class.get_artifacts(artifacts_to_prefetch))
        self.pathspec_index_cache = {}
        self.pathspec_cache = {}
        for ds in datastores:
            self.pathspec_index_cache[ds.pathspec_index] = ds
            self.pathspec_cache[ds.pathspec] = ds

    def get_with_pathspec(self, pathspec):
        return self.pathspec_cache.get(pathspec, None)

    def get_with_pathspec_index(self, pathspec_index):
        return self.pathspec_index_cache.get(pathspec_index, None)

    def __iter__(self):
        for v in self.pathspec_cache.values():
            yield v



#core client classes
from .core import namespace,\
                  get_namespace,\
                  default_namespace,\
                  metadata, \
                  get_metadata, \
                  default_metadata, \
                  Metaflow,\
                  Flow,\
                  Run,\
                  Step,\
                  Task,\
                  DataArtifact



import functools
import json
from multiprocessing import Pool
import os
import tarfile
import shutil
import subprocess
import sys

from metaflow.datatools import S3
from metaflow.metaflow_config import DATASTORE_LOCAL_DIR

from ..env_escape import generate_trampolines, ENV_ESCAPE_PY

from . import CONDA_MAGIC_FILE


def bootstrap_environment(flow_name, env_id):
    setup_conda_manifest(flow_name)
    packages = download_conda_packages(flow_name, env_id)
    install_conda_environment(env_id, packages)

def setup_conda_manifest(flow_name):
    manifest_folder = os.path.join(os.getcwd(), DATASTORE_LOCAL_DIR, flow_name)
    if not os.path.exists(manifest_folder):
        os.makedirs(manifest_folder)
    shutil.move(os.path.join(os.getcwd(), CONDA_MAGIC_FILE), 
        os.path.join(manifest_folder, CONDA_MAGIC_FILE))

def download_conda_packages(flow_name, env_id):
    pkgs_folder = os.path.join(os.getcwd(), 'pkgs')
    if not os.path.exists(pkgs_folder):
        os.makedirs(pkgs_folder)
    manifest_folder = os.path.join(os.getcwd(), DATASTORE_LOCAL_DIR, flow_name)
    with open(os.path.join(manifest_folder, CONDA_MAGIC_FILE)) as f:
        env = json.load(f)[env_id]
        with S3() as s3:
            for pkg in s3.get_many(env['cache_urls']):
                shutil.move(pkg.path, os.path.join(pkgs_folder, os.path.basename(pkg.key)))
        return env['order']

def install_conda_environment(env_id, packages):
    args = [
        'if ! type conda  >/dev/null 2>&1; \
            then wget --no-check-certificate https://repo.anaconda.com/pkgs/misc/conda-execs/conda-latest-linux-64.exe -O conda >/dev/null 2>&1; \
            chmod +x conda; \
            export PATH=$PATH:{0}; fi'.format(os.getcwd()),
        'cd {0}'.format(os.path.join(os.getcwd(), 'pkgs')),
        'conda create --yes --no-default-packages -p {0} --no-deps {1} >/dev/null 2>&1'.format(os.path.join(os.getcwd(), env_id), ' '.join(packages)),
        'cd {0}'.format(os.getcwd())
    ]
    if ENV_ESCAPE_PY is not None:
        cwd = os.getcwd()
        generate_trampolines(cwd)
        # print("Environment escape will use %s as the interpreter" % ENV_ESCAPE_PY)
    else:
        pass
        # print("Could not find a environment escape interpreter")
    os.system(' && '.join(args))

if __name__ == '__main__':
    bootstrap_environment(sys.argv[1], sys.argv[2])


cached_aws_sandbox_creds = None
cached_provider_class = None


class Boto3ClientProvider(object):
    name = "boto3"

    @staticmethod
    def get_client(module, with_error=False, params={}):
        from metaflow.exception import MetaflowException
        from metaflow.metaflow_config import AWS_SANDBOX_ENABLED, \
            AWS_SANDBOX_STS_ENDPOINT_URL, AWS_SANDBOX_API_KEY
        import requests
        try:
            import boto3
            from botocore.exceptions import ClientError
        except (NameError, ImportError):
            raise MetaflowException(
                "Could not import module 'boto3'. Install boto3 first.")

        if AWS_SANDBOX_ENABLED:
            global cached_aws_sandbox_creds
            if cached_aws_sandbox_creds is None:
                # authenticate using STS
                url = "%s/auth/token" % AWS_SANDBOX_STS_ENDPOINT_URL
                headers = {
                    'x-api-key': AWS_SANDBOX_API_KEY
                }
                try:
                    r = requests.get(url, headers=headers)
                    r.raise_for_status()
                    cached_aws_sandbox_creds = r.json()
                except requests.exceptions.HTTPError as e:
                    raise MetaflowException(repr(e))
            if with_error:
                return boto3.session.Session(
                    **cached_aws_sandbox_creds).client(module, **params), ClientError
            return boto3.session.Session(
                **cached_aws_sandbox_creds).client(module, **params)
        if with_error:
            return boto3.client(module, **params), ClientError
        return boto3.client(module, **params)


def get_aws_client(module, with_error=False, params={}):
    global cached_provider_class
    if cached_provider_class is None:
        from metaflow.metaflow_config import DEFAULT_AWS_CLIENT_PROVIDER
        from metaflow.plugins import AWS_CLIENT_PROVIDERS
        for p in AWS_CLIENT_PROVIDERS:
            if p.name == DEFAULT_AWS_CLIENT_PROVIDER:
                cached_provider_class = p
                break
        else:
            raise ValueError("Cannot find AWS Client provider %s"
                             % DEFAULT_AWS_CLIENT_PROVIDER)
    return cached_provider_class.get_client(module, with_error, params)



from __future__ import print_function

import os
import sys
import click
import traceback


# add module to python path if not already present
myDir = os.path.dirname(os.path.abspath(__file__))
parentDir = os.path.split(myDir)[0]
sys.path.insert(0, parentDir)

from metaflow.sidecar_messages import MessageTypes, deserialize
from metaflow.plugins import SIDECARS


class WorkershutdownError(Exception):
    """raised when terminating sidecar"""
    pass


def process_messages(worker):
    while True:
        try:
            msg = sys.stdin.readline().strip()
            if msg:
                parsed_msg = deserialize(msg)
                if parsed_msg.msg_type == MessageTypes.SHUTDOWN:
                    raise WorkershutdownError()
                else:
                    worker.process_message(parsed_msg)
            else:
                raise WorkershutdownError()
        except WorkershutdownError:
            break
        except Exception as e:  # todo handle other possible exceptions gracefully
            print(traceback.format_exc())
            break
    try:
        worker.shutdown()
    except:
        pass


@click.command(help="Initialize workers")
@click.argument('worker-type')
def main(worker_type):

    worker_process = SIDECARS.get(worker_type)

    if worker_process is not None:
        process_messages(worker_process())
    else:
        print("UNRECOGNIZED WORKER: %s" % worker_type, file=sys.stderr)


if __name__ == "__main__":
    main()



import os
import platform
import sys

from .util import get_username, to_unicode
from . import metaflow_version
from metaflow.exception import MetaflowException
from metaflow.mflog import BASH_MFLOG, BASH_SAVE_LOGS
from . import R

version_cache = None


class InvalidEnvironmentException(MetaflowException):
    headline = 'Incompatible environment'


class MetaflowEnvironment(object):
    TYPE = 'local'

    def __init__(self, flow):
        pass

    def init_environment(self, echo):
        """
        Run before any step decorators are initialized.
        """
        pass

    def validate_environment(self, echo):
        """
        Run before any command to validate that we are operating in
        a desired environment.
        """
        pass

    def decospecs(self):
        """
        Environment may insert decorators, equivalent to setting --with
        options on the command line.
        """
        return ()

    def bootstrap_commands(self, step_name):
        """
        A list of shell commands to bootstrap this environment in a remote runtime.
        """
        return []

    def add_to_package(self):
        """
        A list of tuples (file, arcname) to add to the job package.
        `arcname` is an alterative name for the file in the job package.
        """
        return []

    def pylint_config(self):
        """
        Environment may override pylint config.
        """
        return []

    @classmethod
    def get_client_info(cls, flow_name, metadata):
        """
        Environment may customize the information returned to the client about the environment

        Parameters
        ----------
        flow_name : str
            Name of the flow
        metadata : dict
            Metadata information regarding the task

        Returns
        -------
        str : Information printed and returned to the user
        """
        return "Local environment"

    def get_package_commands(self, code_package_url):
        cmds = [BASH_MFLOG,
                "mflog \'Setting up task environment.\'",
                "%s -m pip install awscli click requests boto3 -qqq" 
                    % self._python(),
                "mkdir metaflow",
                "cd metaflow",
                "mkdir .metaflow", # mute local datastore creation log
                "i=0; while [ $i -le 5 ]; do "
                    "mflog \'Downloading code package...\'; "
                    "%s -m awscli s3 cp %s job.tar >/dev/null && \
                        mflog \'Code package downloaded.\' && break; "
                    "sleep 10; i=$((i+1)); "
                "done" % (self._python(), code_package_url),
                "if [ $i -gt 5 ]; then "
                    "mflog \'Failed to download code package from %s "
                    "after 6 tries. Exiting...\' && exit 1; "
                "fi" % code_package_url,
                "TAR_OPTIONS='--warning=no-timestamp' tar xf job.tar",
                "mflog \'Task is starting.\'",
                ]
        return cmds

    def get_environment_info(self):
        global version_cache
        if version_cache is None:
            version_cache = metaflow_version.get_version()

        # note that this dict goes into the code package
        # so variables here should be relatively stable (no
        # timestamps) so the hash won't change all the time
        env = {'platform': platform.system(),
               'username': get_username(),
               'production_token': os.environ.get('METAFLOW_PRODUCTION_TOKEN'),
               'runtime': os.environ.get('METAFLOW_RUNTIME_NAME', 'dev'),
               'app': os.environ.get('APP'),
               'environment_type': self.TYPE,
               'use_r': R.use_r(),
               'python_version': sys.version,
               'python_version_code': '%d.%d.%d' % sys.version_info[:3],
               'metaflow_version': version_cache,
               'script': os.path.basename(os.path.abspath(sys.argv[0]))}
        if R.use_r():
            env['metaflow_r_version'] = R.metaflow_r_version()
            env['r_version'] = R.r_version()
            env['r_version_code'] = R.r_version_code()
        return env

    def executable(self, step_name):
        return self._python()

    def _python(self):
        if R.use_r():
            return "python3"
        else:
            return "python"


import time

from contextlib import contextmanager

@contextmanager
def profile(label, stats_dict=None):
    if stats_dict is None:
        print('PROFILE: %s starting' % label)
    start = time.time()
    yield
    took = int((time.time() - start) * 1000)
    if stats_dict is None:
        print('PROFILE: %s completed in %dms' % (label, took))
    else:
        stats_dict[label] = stats_dict.get(label, 0) + took




from io import BytesIO
from .s3util import aws_retry, get_s3_client

try:
    # python2
    from urlparse import urlparse
except:
    # python3
    from urllib.parse import urlparse

class S3Tail(object):
    def __init__(self, s3url):
        url = urlparse(s3url)
        self.s3, self.ClientError = get_s3_client()
        self._bucket = url.netloc
        self._key = url.path.lstrip('/')
        self._pos = 0
        self._tail = b''

    def clone(self, s3url):
        tail = S3Tail(s3url)
        tail._pos = self._pos
        tail._tail = self._tail
        return tail

    @property
    def bytes_read(self):
        return self._pos

    @property
    def tail(self):
        return self._tail

    def __iter__(self):
        buf = self._fill_buf()
        if buf is not None:
            for line in buf:
                if line.endswith(b'\n'):
                    yield line
                else:
                    self._tail = line
                    break

    @aws_retry
    def _make_range_request(self):
        try:
            return self.s3.get_object(Bucket=self._bucket,
                                      Key=self._key,
                                      Range='bytes=%d-' % self._pos)
        except self.ClientError as err:
            code = err.response['Error']['Code']
            # NOTE we deliberately regard NoSuchKey as an ignorable error.
            # We assume that the file just hasn't appeared in S3 yet.
            if code in ('InvalidRange', 'NoSuchKey'):
                return None
            else:
                raise

    def _fill_buf(self):
        resp = self._make_range_request()
        if resp is None:
            return None
        code = str(resp['ResponseMetadata']['HTTPStatusCode'])
        if code[0] == '2':
            data = resp['Body'].read()
            if data:
                buf = BytesIO(self._tail + data)
                self._pos += len(data)
                self._tail = b''
                return buf
            else:
                return None
        elif code[0] == '5':
            return None
        else:
            raise Exception('Retrieving %s/%s failed: %s' % (self._bucket, self._key, code))




import inspect
import sys
import traceback
from datetime import datetime
from functools import wraps

import click

from . import lint
from . import plugins
from . import parameters
from . import decorators
from . import metaflow_version
from . import namespace
from . import current
from .cli_args import cli_args
from .util import resolve_identity, decompress_list, write_latest_run_id, \
    get_latest_run_id, to_unicode
from .task import MetaflowTask
from .exception import CommandException, MetaflowException
from .graph import FlowGraph
from .datastore import DATASTORES
from .runtime import NativeRuntime
from .package import MetaflowPackage
from .plugins import ENVIRONMENTS, LOGGING_SIDECARS, METADATA_PROVIDERS, MONITOR_SIDECARS
from .metaflow_config import DEFAULT_DATASTORE, DEFAULT_ENVIRONMENT, DEFAULT_EVENT_LOGGER, \
    DEFAULT_METADATA, DEFAULT_MONITOR, DEFAULT_PACKAGE_SUFFIXES
from .metaflow_environment import MetaflowEnvironment
from .pylint_wrapper import PyLint
from .event_logger import EventLogger
from .monitor import Monitor
from .R import use_r, metaflow_r_version
from .mflog import mflog, LOG_SOURCES
from .unbounded_foreach import UBF_CONTROL, UBF_TASK


ERASE_TO_EOL = '\033[K'
HIGHLIGHT = 'red'
INDENT = ' ' * 4

LOGGER_TIMESTAMP = 'magenta'
LOGGER_COLOR = 'green'
LOGGER_BAD_COLOR = 'red'

try:
    # Python 2
    import cPickle as pickle
except ImportError:
    # Python 3
    import pickle


def echo_dev_null(*args, **kwargs):
    pass


def echo_always(line, **kwargs):
    kwargs['err'] = kwargs.get('err', True)
    if kwargs.pop('indent', None):
        line = '\n'.join(INDENT + x for x in line.splitlines())
    if 'nl' not in kwargs or kwargs['nl']:
        line += ERASE_TO_EOL
    top = kwargs.pop('padding_top', None)
    bottom = kwargs.pop('padding_bottom', None)
    highlight = kwargs.pop('highlight', HIGHLIGHT)
    if top:
        click.secho(ERASE_TO_EOL, **kwargs)

    hl_bold = kwargs.pop('highlight_bold', True)
    nl = kwargs.pop('nl', True)
    fg = kwargs.pop('fg', None)
    bold = kwargs.pop('bold', False)
    kwargs['nl'] = False
    hl = True
    nobold = kwargs.pop('no_bold', False)
    if nobold:
        click.secho(line, **kwargs)
    else:
        for span in line.split('*'):
            if hl:
                hl = False
                kwargs['fg'] = fg
                kwargs['bold'] = bold
                click.secho(span, **kwargs)
            else:
                hl = True
                kwargs['fg'] = highlight
                kwargs['bold'] = hl_bold
                click.secho(span, **kwargs)
    if nl:
        kwargs['nl'] = True
        click.secho('', **kwargs)
    if bottom:
        click.secho(ERASE_TO_EOL, **kwargs)


def logger(body='',
           system_msg=False,
           head='',
           bad=False,
           timestamp=True):
    if timestamp:
        if timestamp is True:
            dt = datetime.now()
        else:
            dt = timestamp
        tstamp = dt.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
        click.secho(tstamp + ' ', fg=LOGGER_TIMESTAMP, nl=False)
    if head:
        click.secho(head, fg=LOGGER_COLOR, nl=False)
    click.secho(body,
                bold=system_msg,
                fg=LOGGER_BAD_COLOR if bad else None)


@click.group()
def cli(ctx):
    pass


@cli.command(help='Check that the flow is valid (default).')
@click.option('--warnings/--no-warnings',
              default=False,
              show_default=True,
              help='Show all Pylint warnings, not just errors.')
@click.pass_obj
def check(obj, warnings=False):
    _check(obj.graph, obj.flow, obj.environment, pylint=obj.pylint, warnings=warnings)
    fname = inspect.getfile(obj.flow.__class__)
    echo("\n*'{cmd} show'* shows a description of this flow.\n"
         "*'{cmd} run'* runs the flow locally.\n"
         "*'{cmd} help'* shows all available commands and options.\n"
         .format(cmd=fname), highlight='magenta', highlight_bold=False)


@cli.command(help='Show structure of the flow.')
@click.pass_obj
def show(obj):
    echo_always('\n%s' % obj.graph.doc)
    for _, node in sorted((n.func_lineno, n) for n in obj.graph):
        echo_always('\nStep *%s*' % node.name, err=False)
        echo_always(node.doc if node.doc else '?', indent=True, err=False)
        if node.type != 'end':
            echo_always('*=>* %s' % ', '.join('*%s*' % n for n in node.out_funcs),
                        indent=True,
                        highlight='magenta',
                        highlight_bold=False,
                        err=False)
    echo_always('')


@cli.command(help='Show all available commands.')
@click.pass_context
def help(ctx):
    print(ctx.parent.get_help())


@cli.command(help='Output internal state of the flow graph.')
@click.pass_obj
def output_raw(obj):
    echo('Internal representation of the flow:',
         fg='magenta',
         bold=False)
    echo_always(str(obj.graph), err=False)


@cli.command(help='Visualize the flow with Graphviz.')
@click.pass_obj
def output_dot(obj):
    echo('Visualizing the flow as a GraphViz graph',
         fg='magenta',
         bold=False)
    echo("Try piping the output to 'dot -Tpng -o graph.png' to produce "
         "an actual image.", indent=True)
    echo_always(obj.graph.output_dot(), err=False)


@cli.command(help='Get data artifacts of a task or all tasks in a step. '
                  'The format for input-path is either <run_id>/<step_name> or '
                  '<run_id>/<step_name>/<task_id>.')
@click.argument('input-path')
@click.option('--private/--no-private',
              default=False,
              show_default=True,
              help='Show also private attributes.')
@click.option('--max-value-size',
              default=1000,
              show_default=True,
              type=int,
              help='Show only values that are smaller than this number. '
                   'Set to 0 to see only keys.')
@click.option('--include',
              type=str,
              default='',
              help='Include only artifacts in the given comma-separated list.')
@click.option('--file',
              type=str,
              default=None,
              help='Serialize artifacts in the given file.')
@click.pass_obj
def dump(obj,
         input_path,
         private=None,
         max_value_size=None,
         include=None,
         file=None):

    output = {}
    kwargs = {'show_private': private,
              'max_value_size': max_value_size,
              'include': {t for t in include.split(',') if t}}

    if obj.datastore.datastore_root is None:
        obj.datastore.datastore_root = obj.datastore.get_datastore_root_from_config(
            obj.echo, create_on_absent=False)
    if obj.datastore.datastore_root is None:
        raise CommandException(
            "Could not find the location of the datastore -- did you correctly set the "
            "METAFLOW_DATASTORE_SYSROOT_%s environment variable" % (obj.datastore.TYPE).upper())

    # Pathspec can either be run_id/step_name or run_id/step_name/task_id.
    parts = input_path.split('/')
    if len(parts) == 2:
        run_id, step_name = parts
        task_id = None
    elif len(parts) == 3:
        run_id, step_name, task_id = parts
    else:
        raise CommandException("input_path should either be run_id/step_name"
                               "or run_id/step_name/task_id")

    from metaflow.datastore.datastore_set import MetaflowDatastoreSet

    datastore_set = MetaflowDatastoreSet(
                        obj.datastore,
                        obj.flow.name,
                        run_id,
                        steps=[step_name],
                        metadata=obj.metadata,
                        monitor=obj.monitor,
                        event_logger=obj.event_logger,
                        prefetch_data_artifacts=kwargs.get('include'))
    if task_id:
        ds_list = [datastore_set.get_with_pathspec(input_path)]
    else:
        ds_list = list(datastore_set) # get all tasks

    for ds in ds_list:
        echo('Dumping output of run_id=*{run_id}* '
             'step=*{step}* task_id=*{task_id}*'.format(run_id=ds.run_id,
                                                        step=ds.step_name,
                                                        task_id=ds.task_id),
             fg='magenta')

        if file is None:
            echo_always(ds.format(**kwargs),
                        highlight='green',
                        highlight_bold=False,
                        err=False)
        else:
            output[ds.pathspec] = ds.to_dict(**kwargs)

    if file is not None:
        with open(file, 'wb') as f:
            pickle.dump(output, f, protocol=pickle.HIGHEST_PROTOCOL)
        echo('Artifacts written to *%s*' % file)


@cli.command(help='Show stdout/stderr produced by a task or all tasks in a step. '
                  'The format for input-path is either <run_id>/<step_name> or '
                  '<run_id>/<step_name>/<task_id>.')
@click.argument('input-path')
@click.option('--stdout/--no-stdout',
              default=False,
              show_default=True,
              help='Show stdout of the task.')
@click.option('--stderr/--no-stderr',
              default=False,
              show_default=True,
              help='Show stderr of the task.')
@click.option('--both/--no-both',
              default=True,
              show_default=True,
              help='Show both stdout and stderr of the task.')
@click.option('--timestamps/--no-timestamps',
              default=False,
              show_default=True,
              help='Show timestamps.')
@click.pass_obj
def logs(obj,
         input_path,
         stdout=None,
         stderr=None,
         both=None,
         timestamps=False):
    types = set()
    if stdout:
        types.add('stdout')
        both = False
    if stderr:
        types.add('stderr')
        both = False
    if both:
        types.update(('stdout', 'stderr'))

    streams = list(sorted(types, reverse=True))

    # Pathspec can either be run_id/step_name or run_id/step_name/task_id.
    parts = input_path.split('/')
    if len(parts) == 2:
        run_id, step_name = parts
        task_id = None
    elif len(parts) == 3:
        run_id, step_name, task_id = parts
    else:
        raise CommandException("input_path should either be run_id/step_name "
                               "or run_id/step_name/task_id")

    if obj.datastore.datastore_root is None:
        obj.datastore.datastore_root = obj.datastore.get_datastore_root_from_config(
            obj.echo, create_on_absent=False)
    if obj.datastore.datastore_root is None:
        raise CommandException(
            "Could not find the location of the datastore -- did you correctly set the "
            "METAFLOW_DATASTORE_SYSROOT_%s environment variable" % (obj.datastore.TYPE).upper())

    if task_id:
        ds_list = [obj.datastore(obj.flow.name,
                                 run_id=run_id,
                                 step_name=step_name,
                                 task_id=task_id,
                                 mode='r',
                                 allow_unsuccessful=True)]
    else:
        from metaflow.datastore.datastore_set import MetaflowDatastoreSet
        datastore_set = MetaflowDatastoreSet(
                            obj.datastore,
                            obj.flow.name,
                            run_id,
                            steps=[step_name],
                            metadata=obj.metadata,
                            monitor=obj.monitor,
                            event_logger=obj.event_logger)
        # get all successful tasks
        ds_list = list(datastore_set)

    if ds_list:
        def echo_unicode(line, **kwargs):
            click.secho(line.decode('UTF-8', errors='replace'), **kwargs)

        # old style logs are non mflog-style logs
        maybe_old_style = True
        for ds in ds_list:
            echo('Dumping logs of run_id=*{run_id}* '
                 'step=*{step}* task_id=*{task_id}*'.format(run_id=ds.run_id,
                                                            step=ds.step_name,
                                                            task_id=ds.task_id),
                 fg='magenta')

            for stream in streams:
                echo(stream, bold=True)
                logs = ds.load_logs(LOG_SOURCES, stream)
                if any(data for _, data in logs):
                    # attempt to read new, mflog-style logs
                    for line in mflog.merge_logs([blob for _, blob in logs]):
                        if timestamps:
                            ts = mflog.utc_to_local(line.utc_tstamp)
                            tstamp = ts.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
                            click.secho(tstamp + ' ',
                                        fg=LOGGER_TIMESTAMP,
                                        nl=False)
                        echo_unicode(line.msg)
                    maybe_old_style = False
                elif maybe_old_style:
                    # if they are not available, we may be looking at
                    # a legacy run (unless we have seen new-style data already
                    # for another stream). This return an empty string if
                    # nothing is found
                    log = ds.load_log_legacy(stream)
                    if log and timestamps:
                        raise CommandException("We can't show --timestamps for "
                                               "old runs. Sorry!")
                    echo_unicode(log, nl=False)

    elif len(parts) == 2:
        # TODO if datastore provided a way to find unsuccessful task IDs, we
        # could make handle this case automatically
        raise CommandException("Successful tasks were not found at the given "
                               "path. You can see logs for unsuccessful tasks "
                               "by giving an exact task ID using the "
                               "run_id/step_name/task_id format.")


# TODO - move step and init under a separate 'internal' subcommand

@cli.command(help="Internal command to execute a single task.")
@click.argument('step-name')
@click.option('--run-id',
              default=None,
              required=True,
              help='ID for one execution of all steps in the flow.')
@click.option('--task-id',
              default=None,
              required=True,
              show_default=True,
              help='ID for this instance of the step.')
@click.option('--input-paths',
              help='A comma-separated list of pathspecs '
                   'specifying inputs for this step.')
@click.option('--split-index',
              type=int,
              default=None,
              show_default=True,
              help='Index of this foreach split.')
@click.option('--tag',
              'opt_tag',
              multiple=True,
              default=None,
              help="Annotate this run with the given tag. You can specify "
                   "this option multiple times to attach multiple tags in "
                   "the task.")
@click.option('--namespace',
              'opt_namespace',
              default=None,
              help="Change namespace from the default (your username) to "
                   "the specified tag.")
@click.option('--retry-count',
              default=0,
              help="How many times we have attempted to run this task.")
@click.option('--max-user-code-retries',
              default=0,
              help="How many times we should attempt running the user code.")
@click.option('--clone-only',
              default=None,
              help="Pathspec of the origin task for this task to clone. Do "
              "not execute anything.")
@click.option('--clone-run-id',
              default=None,
              help="Run id of the origin flow, if this task is part of a flow "
              "being resumed.")
@click.option('--with',
              'decospecs',
              multiple=True,
              help="Add a decorator to this task. You can specify this "
              "option multiple times to attach multiple decorators "
              "to this task.")
@click.option('--ubf-context',
              default='none',
              type=click.Choice(['none', UBF_CONTROL, UBF_TASK]),
              help="Provides additional context if this task is of type "
              "unbounded foreach.")
@click.pass_context
def step(ctx,
         step_name,
         opt_tag=None,
         run_id=None,
         task_id=None,
         input_paths=None,
         split_index=None,
         opt_namespace=None,
         retry_count=None,
         max_user_code_retries=None,
         clone_only=None,
         clone_run_id=None,
         decospecs=None,
         ubf_context=None):
    if opt_namespace is not None:
        namespace(opt_namespace or None)

    func = None
    try:
        func = getattr(ctx.obj.flow, step_name)
    except:
        raise CommandException("Step *%s* doesn't exist." % step_name)
    if not func.is_step:
        raise CommandException("Function *%s* is not a step." % step_name)
    echo('Executing a step, *%s*' % step_name,
         fg='magenta',
         bold=False)

    if decospecs:
        decorators._attach_decorators_to_step(func, decospecs)

    ctx.obj.datastore.datastore_root = ctx.obj.datastore_root
    if ctx.obj.datastore.datastore_root is None:
        ctx.obj.datastore.datastore_root = \
            ctx.obj.datastore.get_datastore_root_from_config(ctx.obj.echo)

    step_kwargs = ctx.params
    # Remove argument `step_name` from `step_kwargs`.
    step_kwargs.pop('step_name', None)
    # Remove `opt_*` prefix from (some) option keys.
    step_kwargs = dict([(k[4:], v) if k.startswith('opt_') else (k, v)
                         for k, v in step_kwargs.items()])
    cli_args._set_step_kwargs(step_kwargs)

    ctx.obj.metadata.add_sticky_tags(tags=opt_tag)
    paths = decompress_list(input_paths) if input_paths else []

    task = MetaflowTask(ctx.obj.flow,
                        ctx.obj.datastore,
                        ctx.obj.metadata,
                        ctx.obj.environment,
                        ctx.obj.echo,
                        ctx.obj.event_logger,
                        ctx.obj.monitor,
                        ubf_context)
    if clone_only:
        task.clone_only(step_name,
                        run_id,
                        task_id,
                        clone_only)
    else:
        task.run_step(step_name,
                      run_id,
                      task_id,
                      clone_run_id,
                      paths,
                      split_index,
                      retry_count,
                      max_user_code_retries)

    echo('Success', fg='green', bold=True, indent=True)

@parameters.add_custom_parameters(deploy_mode=False)
@cli.command(help="Internal command to initialize a run.")
@click.option('--run-id',
              default=None,
              required=True,
              help='ID for one execution of all steps in the flow.')
@click.option('--task-id',
              default=None,
              required=True,
              help='ID for this instance of the step.')
@click.option('--tag',
              'tags',
              multiple=True,
              default=None,
              help="Tags for this instance of the step.")
@click.pass_obj
def init(obj, run_id=None, task_id=None, tags=None, **kwargs):
    # init is a separate command instead of an option in 'step'
    # since we need to capture user-specified parameters with
    # @add_custom_parameters. Adding custom parameters to 'step'
    # is not desirable due to the possibility of name clashes between
    # user-specified parameters and our internal options. Note that
    # user-specified parameters are often defined as environment
    # variables.

    if obj.datastore.datastore_root is None:
        obj.datastore.datastore_root = \
            obj.datastore.get_datastore_root_from_config(obj.echo)

    obj.metadata.add_sticky_tags(tags=tags)

    runtime = NativeRuntime(obj.flow,
                            obj.graph,
                            obj.datastore,
                            obj.metadata,
                            obj.environment,
                            obj.package,
                            obj.logger,
                            obj.entrypoint,
                            obj.event_logger,
                            obj.monitor,
                            run_id=run_id)
    parameters.set_parameters(obj.flow, kwargs)
    runtime.persist_parameters(task_id=task_id)

def common_run_options(func):
    @click.option('--tag',
                  'tags',
                  multiple=True,
                  default=None,
                  help="Annotate this run with the given tag. You can specify "
                       "this option multiple times to attach multiple tags in "
                       "the run.")
    @click.option('--max-workers',
                  default=16,
                  show_default=True,
                  help='Maximum number of parallel processes.')
    @click.option('--max-num-splits',
                  default=100,
                  show_default=True,
                  help='Maximum number of splits allowed in a foreach. This '
                       'is a safety check preventing bugs from triggering '
                       'thousands of steps inadvertently.')
    @click.option('--max-log-size',
                  default=10,
                  show_default=True,
                  help='Maximum size of stdout and stderr captured in '
                       'megabytes. If a step outputs more than this to '
                       'stdout/stderr, its output will be truncated.')
    @click.option('--with',
                  'decospecs',
                  multiple=True,
                  help="Add a decorator to all steps. You can specify this "
                       "option multiple times to attach multiple decorators "
                       "in steps.")
    @click.option('--run-id-file',
                  default=None,
                  show_default=True,
                  type=str,
                  help="Write the ID of this run to the file specified.")
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper


@click.option('--origin-run-id',
              default=None,
              help="ID of the run that should be resumed. By default, the "
                   "last run executed locally.")
@click.argument('step-to-rerun',
                required=False)
@cli.command(help='Resume execution of a previous run of this flow.')
@common_run_options
@click.pass_obj
def resume(obj,
           tags=None,
           step_to_rerun=None,
           origin_run_id=None,
           max_workers=None,
           max_num_splits=None,
           max_log_size=None,
           decospecs=None,
           run_id_file=None):

    before_run(obj, tags, decospecs + obj.environment.decospecs())

    if origin_run_id is None:
        origin_run_id = get_latest_run_id(obj.echo, obj.flow.name)
        if origin_run_id is None:
            raise CommandException("A previous run id was not found. Specify --origin-run-id.")

    if step_to_rerun is None:
        clone_steps = set()
    else:
        # validate step name
        if step_to_rerun not in obj.graph.nodes:
            raise CommandException(
                "invalid step name {0} specified, must be step present in "
                "current form of execution graph. Valid step names include: {1}"
                .format(
                    step_to_rerun,
                    ",".join(list(obj.graph.nodes.keys()))))
        clone_steps = {step_to_rerun}

    runtime = NativeRuntime(obj.flow,
                            obj.graph,
                            obj.datastore,
                            obj.metadata,
                            obj.environment,
                            obj.package,
                            obj.logger,
                            obj.entrypoint,
                            obj.event_logger,
                            obj.monitor,
                            clone_run_id=origin_run_id,
                            clone_steps=clone_steps,
                            max_workers=max_workers,
                            max_num_splits=max_num_splits,
                            max_log_size=max_log_size * 1024 * 1024)
    runtime.persist_parameters()
    runtime.execute()

    write_run_id(run_id_file, runtime.run_id)


@parameters.add_custom_parameters(deploy_mode=True)
@cli.command(help='Run the workflow locally.')
@common_run_options
@click.option('--namespace',
              'user_namespace',
              default=None,
              help="Change namespace from the default (your username) to "
                   "the specified tag. Note that this option does not alter "
                   "tags assigned to the objects produced by this run, just "
                   "what existing objects are visible in the client API. You "
                   "can enable the global namespace with an empty string."
                   "--namespace=")
@click.pass_obj
def run(obj,
        tags=None,
        max_workers=None,
        max_num_splits=None,
        max_log_size=None,
        decospecs=None,
        run_id_file=None,
        user_namespace=None,
        **kwargs):

    if user_namespace is not None:
        namespace(user_namespace or None)
    before_run(obj, tags, decospecs + obj.environment.decospecs())

    runtime = NativeRuntime(obj.flow,
                            obj.graph,
                            obj.datastore,
                            obj.metadata,
                            obj.environment,
                            obj.package,
                            obj.logger,
                            obj.entrypoint,
                            obj.event_logger,
                            obj.monitor,
                            max_workers=max_workers,
                            max_num_splits=max_num_splits,
                            max_log_size=max_log_size * 1024 * 1024)
    write_latest_run_id(obj, runtime.run_id)
    write_run_id(run_id_file, runtime.run_id)

    parameters.set_parameters(obj.flow, kwargs)
    runtime.persist_parameters()
    runtime.execute()


def write_run_id(run_id_file, run_id):
    if run_id_file is not None:
        with open(run_id_file, 'w') as f:
            f.write(str(run_id))


def before_run(obj, tags, decospecs):
    # There's a --with option both at the top-level and for the run
    # subcommand. Why?
    #
    # "run --with shoes" looks so much better than "--with shoes run".
    # This is a very common use case of --with.
    #
    # A downside is that we need to have the following decorators handling
    # in two places in this module and we need to make sure that
    # _init_step_decorators doesn't get called twice.
    if decospecs:
        decorators._attach_decorators(obj.flow, decospecs)
        obj.graph = FlowGraph(obj.flow.__class__)
    obj.check(obj.graph, obj.flow, obj.environment, pylint=obj.pylint)
    #obj.environment.init_environment(obj.logger)

    if obj.datastore.datastore_root is None:
        obj.datastore.datastore_root = \
            obj.datastore.get_datastore_root_from_config(obj.echo)

    decorators._init_step_decorators(
        obj.flow, obj.graph, obj.environment, obj.datastore, obj.logger)
    obj.metadata.add_sticky_tags(tags=tags)

    # Package working directory only once per run.
    # We explicitly avoid doing this in `start` since it is invoked for every
    # step in the run.
    obj.package = MetaflowPackage(obj.flow,
                                  obj.environment,
                                  obj.echo,
                                  obj.package_suffixes)


@cli.command(help='Print the Metaflow version')
@click.pass_obj
def version(obj):
    echo_always(obj.version)

@decorators.add_decorator_options
@click.command(cls=click.CommandCollection,
               sources=[cli] + plugins.get_plugin_cli(),
               invoke_without_command=True)
@click.option('--quiet/--not-quiet',
              show_default=True,
              default=False,
              help='Suppress unnecessary messages')
@click.option('--metadata',
              default=DEFAULT_METADATA,
              show_default=True,
              type=click.Choice([m.TYPE for m in METADATA_PROVIDERS]),
              help='Metadata service type')
@click.option('--environment',
              default=DEFAULT_ENVIRONMENT,
              show_default=True,
              type=click.Choice(['local'] + [m.TYPE for m in ENVIRONMENTS]),
              help='Execution environment type')
@click.option('--datastore',
              default=DEFAULT_DATASTORE,
              show_default=True,
              type=click.Choice(DATASTORES),
              help='Data backend type')
@click.option('--datastore-root',
              help='Root path for datastore')
@click.option('--package-suffixes',
              help='A comma-separated list of file suffixes to include '
                   'in the code package.',
              default=DEFAULT_PACKAGE_SUFFIXES,
              show_default=True)
@click.option('--with',
              'decospecs',
              multiple=True,
              help="Add a decorator to all steps. You can specify this option "
                   "multiple times to attach multiple decorators in steps.")
@click.option('--pylint/--no-pylint',
              default=True,
              show_default=True,
              help='Run Pylint on the flow if pylint is installed.')
@click.option('--coverage',
              is_flag=True,
              default=False,
              show_default=True,
              help='Measure code coverage using coverage.py.')
@click.option('--event-logger',
              default=DEFAULT_EVENT_LOGGER,
              show_default=True,
              type=click.Choice(LOGGING_SIDECARS),
              help='type of event logger used')
@click.option('--monitor',
              default=DEFAULT_MONITOR,
              show_default=True,
              type=click.Choice(MONITOR_SIDECARS),
              help='Monitoring backend type')
@click.pass_context
def start(ctx,
          quiet=False,
          metadata=None,
          environment=None,
          datastore=None,
          datastore_root=None,
          decospecs=None,
          package_suffixes=None,
          pylint=None,
          coverage=None,
          event_logger=None,
          monitor=None,
          **deco_options):
    global echo
    if quiet:
        echo = echo_dev_null
    else:
        echo = echo_always

    ctx.obj.version = metaflow_version.get_version()
    version = ctx.obj.version
    if use_r():
        version = metaflow_r_version()

    echo('Metaflow %s' % version, fg='magenta', bold=True, nl=False)
    echo(" executing *%s*" % ctx.obj.flow.name, fg='magenta', nl=False)
    echo(" for *%s*" % resolve_identity(), fg='magenta')

    if coverage:
        from coverage import Coverage
        cov = Coverage(data_suffix=True,
                       auto_data=True,
                       source=['metaflow'],
                       branch=True)
        cov.start()

    cli_args._set_top_kwargs(ctx.params)
    ctx.obj.echo = echo
    ctx.obj.echo_always = echo_always
    ctx.obj.graph = FlowGraph(ctx.obj.flow.__class__)
    ctx.obj.logger = logger
    ctx.obj.check = _check
    ctx.obj.pylint = pylint
    ctx.obj.top_cli = cli
    ctx.obj.package_suffixes = package_suffixes.split(',')
    ctx.obj.reconstruct_cli = _reconstruct_cli

    ctx.obj.event_logger = EventLogger(event_logger)

    ctx.obj.environment = [e for e in ENVIRONMENTS + [MetaflowEnvironment]
                           if e.TYPE == environment][0](ctx.obj.flow)
    ctx.obj.environment.validate_environment(echo)

    ctx.obj.datastore = DATASTORES[datastore]
    ctx.obj.datastore_root = datastore_root

    # It is important to initialize flow decorators early as some of the
    # things they provide may be used by some of the objects initialize after.
    decorators._init_flow_decorators(ctx.obj.flow,
                                     ctx.obj.graph,
                                     ctx.obj.environment,
                                     ctx.obj.datastore,
                                     ctx.obj.logger,
                                     echo,
                                     deco_options)

    ctx.obj.monitor = Monitor(monitor, ctx.obj.environment, ctx.obj.flow.name)
    ctx.obj.monitor.start()

    ctx.obj.metadata = [m for m in METADATA_PROVIDERS
                        if m.TYPE == metadata][0](ctx.obj.environment,
                                                  ctx.obj.flow,
                                                  ctx.obj.event_logger,
                                                  ctx.obj.monitor)
    ctx.obj.datastore = DATASTORES[datastore]

    if datastore_root is None:
        datastore_root = \
          ctx.obj.datastore.get_datastore_root_from_config(ctx.obj.echo)
    ctx.obj.datastore_root = ctx.obj.datastore.datastore_root = datastore_root

    if decospecs:
        decorators._attach_decorators(ctx.obj.flow, decospecs)

    # initialize current and parameter context for deploy-time parameters
    current._set_env(flow_name=ctx.obj.flow.name, is_running=False)
    parameters.set_parameter_context(ctx.obj.flow.name,
                                        ctx.obj.echo,
                                        ctx.obj.datastore)

    if ctx.invoked_subcommand not in ('run', 'resume'):
        # run/resume are special cases because they can add more decorators with --with,
        # so they have to take care of themselves.
        decorators._attach_decorators(
            ctx.obj.flow, ctx.obj.environment.decospecs())
        decorators._init_step_decorators(
            ctx.obj.flow, ctx.obj.graph, ctx.obj.environment, ctx.obj.datastore, ctx.obj.logger)
        #TODO (savin): Enable lazy instantiation of package
        ctx.obj.package = None
    if ctx.invoked_subcommand is None:
        ctx.invoke(check)


def _reconstruct_cli(params):
    for k, v in params.items():
        if v:
            if k == 'decospecs':
                k = 'with'
            k = k.replace('_', '-')
            if not isinstance(v, tuple):
                v = [v]
            for value in v:
                yield '--%s' % k
                if not isinstance(value, bool):
                    yield str(value)


def _check(graph, flow, environment, pylint=True, warnings=False, **kwargs):
    echo("Validating your flow...", fg='magenta', bold=False)
    linter = lint.linter
    # TODO set linter settings
    linter.run_checks(graph, **kwargs)
    echo('The graph looks good!', fg='green', bold=True, indent=True)
    if pylint:
        echo("Running pylint...", fg='magenta', bold=False)
        fname = inspect.getfile(flow.__class__)
        pylint = PyLint(fname)
        if pylint.has_pylint():
            pylint_is_happy, pylint_exception_msg = pylint.run(
                        warnings=warnings,
                        pylint_config=environment.pylint_config(),
                        logger=echo_always)

            if pylint_is_happy:
                echo('Pylint is happy!',
                     fg='green',
                     bold=True,
                     indent=True)
            else:
                echo("Pylint couldn't analyze your code.\n\tPylint exception: %s"
                     % pylint_exception_msg,
                     fg='red',
                     bold=True,
                     indent=True)
                echo("Skipping Pylint checks.",
                     fg='red',
                     bold=True,
                     indent=True)
        else:
            echo("Pylint not found, so extra checks are disabled.",
                 fg='green',
                 indent=True,
                 bold=False)


def print_metaflow_exception(ex):
    echo_always(ex.headline, indent=True, nl=False, bold=True)
    if ex.line_no is None:
        echo_always(':')
    else:
        echo_always(' on line %d:' % ex.line_no, bold=True)
    echo_always(ex.message, indent=True, bold=False, padding_bottom=True)


def print_unknown_exception(ex):
    echo_always('Internal error', indent=True, bold=True)
    echo_always(traceback.format_exc(), highlight=None, highlight_bold=False)


class CliState(object):
    def __init__(self, flow):
        self.flow = flow

def main(flow, args=None, handle_exceptions=True, entrypoint=None):
    # Ignore warning(s) and prevent spamming the end-user.
    # TODO: This serves as a short term workaround for RuntimeWarning(s) thrown
    # in py3.8 related to log buffering (bufsize=1).
    import warnings
    warnings.filterwarnings('ignore')
    if entrypoint is None:
        entrypoint = [sys.executable, sys.argv[0]]

    state = CliState(flow)
    state.entrypoint = entrypoint

    try:
        if args is None:
            start(auto_envvar_prefix='METAFLOW', obj=state)
        else:
            try:
                start.main(args=args,
                           obj=state,
                           auto_envvar_prefix='METAFLOW')
            except SystemExit as e:
                return e.code
    except MetaflowException as x:
        if handle_exceptions:
            print_metaflow_exception(x)
            sys.exit(1)
        else:
            raise
    except Exception as x:
        if handle_exceptions:
            print_unknown_exception(x)
            sys.exit(1)
        else:
            raise



import os
import requests
from metaflow.metaflow_config import SFN_DYNAMO_DB_TABLE


class DynamoDbClient(object):

    def __init__(self):
        from ..aws_client import get_aws_client
        self._client = get_aws_client(
            'dynamodb',
            params={'region_name': self._get_instance_region()})
        self.name = SFN_DYNAMO_DB_TABLE

    def save_foreach_cardinality(self, 
                                 foreach_split_task_id, 
                                 foreach_cardinality,
                                 ttl):
        return self._client.put_item(
            TableName = self.name,
            Item = {
                'pathspec': {
                    'S': foreach_split_task_id
                },
                'for_each_cardinality': {
                    'NS': list(map(str, range(foreach_cardinality)))
                },
                'ttl': {
                    'N': str(ttl)
                }
            }
        )

    def save_parent_task_id_for_foreach_join(self, 
                                             foreach_split_task_id, 
                                             foreach_join_parent_task_id):
        return self._client.update_item(
            TableName = self.name,
            Key = {
                'pathspec': {
                    'S': foreach_split_task_id
                }
            },
            UpdateExpression = 'ADD parent_task_ids_for_foreach_join :val',
            ExpressionAttributeValues = {
                ':val': {
                    'SS': [foreach_join_parent_task_id] 
                }
            }
        )

    def get_parent_task_ids_for_foreach_join(self,
                                             foreach_split_task_id):
        response = self._client.get_item(
                TableName = self.name,
                Key = {
                    'pathspec': {
                        'S': foreach_split_task_id
                    }
                },
                ProjectionExpression = 'parent_task_ids_for_foreach_join',
                ConsistentRead = True
            )
        return response['Item']['parent_task_ids_for_foreach_join']['SS']

    def _get_instance_region(self):
        region = os.environ.get('AWS_REGION')
        # region is available as an env variable in AWS Fargate but not in EC2
        if region is not None:
            return region
        metadata_url = "http://169.254.169.254/latest/meta-data/placement/availability-zone/"
        r = requests.get(
            url = metadata_url
        )

        if r.status_code != 200:
            raise RuntimeError("Failed to query instance metadata. Url [%s]" % metadata_url +
                        " Error code [%s]" % str(r.status_code))

        return r.text[:-1]


import base64
import click
from hashlib import sha1
import json
import re
from distutils.version import LooseVersion

from metaflow import current, decorators, parameters, JSONType
from metaflow.metaflow_config import SFN_STATE_MACHINE_PREFIX
from metaflow.exception import MetaflowException, MetaflowInternalError
from metaflow.datastore.datastore import TransformableObject
from metaflow.package import MetaflowPackage
from metaflow.plugins import BatchDecorator
from metaflow.util import get_username, to_bytes, to_unicode

from .step_functions import StepFunctions
from .production_token import load_token, store_token, new_token

VALID_NAME = re.compile('[^a-zA-Z0-9_\-\.]')

class IncorrectProductionToken(MetaflowException):
    headline = "Incorrect production token"

class IncorrectMetadataServiceVersion(MetaflowException):
    headline = "Incorrect version for metaflow service"

class StepFunctionsStateMachineNameTooLong(MetaflowException):
    headline = "AWS Step Functions state machine name too long"

@click.group()
def cli():
    pass

@cli.group(help="Commands related to AWS Step Functions.")
@click.option('--name',
              default=None,
              type=str,
              help="State Machine name. The flow name is used instead "
                   "if this option is not specified")
@click.pass_obj
def step_functions(obj,
                   name=None):
    obj.check(obj.graph, obj.flow, obj.environment, pylint=obj.pylint)
    obj.state_machine_name, obj.token_prefix, obj.is_project = \
                                      resolve_state_machine_name(obj, name)

@step_functions.command(help="Deploy a new version of this workflow to "
                    "AWS Step Functions.")
@click.option('--authorize',
              default=None,
              help="Authorize using this production token. You need this "
                   "when you are re-deploying an existing flow for the first "
                   "time. The token is cached in METAFLOW_HOME, so you only "
                   "need to specify this once.")
@click.option('--generate-new-token',
              is_flag=True,
              help="Generate a new production token for this flow. "
                   "This will move the production flow to a new "
                   "namespace.")
@click.option('--new-token',
              'given_token',
              default=None,
              help="Use the given production token for this flow. "
                   "This will move the production flow to the given "
                   "namespace.")
@click.option('--tag',
              'tags',
              multiple=True,
              default=None,
              help="Annotate all objects produced by AWS Step Functions runs "
                   "with the given tag. You can specify this option multiple "
                   "times to attach multiple tags.")
@click.option('--namespace',
              'user_namespace',
              default=None,
              help="Change the namespace from the default (production token) "
                   "to the given tag. See run --help for more information.")
@click.option('--only-json',
              is_flag=True,
              default=False,
              help="Only print out JSON sent to AWS Step Functions. Do not "
                   "deploy anything.")
@click.option('--max-workers',
              default=100,
              show_default=True,
              help="Maximum number of parallel processes.")
@click.option('--workflow-timeout',
              default=None,
              type=int,
              help="Workflow timeout in seconds.")
@click.option('--log-execution-history',
              is_flag=True,
              help="Log AWS Step Functions execution history to AWS CloudWatch "
                   "Logs log group.")
@click.pass_obj
def create(obj,
           tags=None,
           user_namespace=None,
           only_json=False,
           authorize=None,
           generate_new_token=False,
           given_token=None,
           max_workers=None,
           workflow_timeout=None,
           log_execution_history=False):
    obj.echo("Deploying *%s* to AWS Step Functions..." % obj.state_machine_name,
             bold=True)

    check_metadata_service_version(obj)

    token = resolve_token(obj.state_machine_name,
                          obj.token_prefix,
                          obj,
                          authorize,
                          given_token,
                          generate_new_token,
                          obj.is_project)

    flow = make_flow(obj,
                     token,
                     obj.state_machine_name,
                     tags,
                     user_namespace,
                     max_workers,
                     workflow_timeout,
                     obj.is_project)

    if only_json:
        obj.echo_always(flow.to_json(), err=False, no_bold=True)
    else:
        flow.deploy(log_execution_history)
        obj.echo("State Machine *{state_machine}* "
                 "for flow *{name}* pushed to "
                 "AWS Step Functions successfully.\n"
                    .format(state_machine=obj.state_machine_name,
                            name=current.flow_name), 
                 bold=True)
        if obj._is_state_machine_name_hashed:
            obj.echo("Note that the flow was deployed with a truncated name "
                     "due to a length limit on AWS Step Functions. The "
                     "original long name is stored in task metadata.\n")
        flow.schedule()
        obj.echo("What will trigger execution of the workflow:", bold=True)
        obj.echo(flow.trigger_explanation(), indent=True)
        

def check_metadata_service_version(obj):
    metadata = obj.metadata
    version = metadata.version()
    if version == 'local':
        return
    elif version is not None and LooseVersion(version) >= LooseVersion('2.0.2'):
        # Metaflow metadata service needs to be at least at version 2.0.2
        return
    else:
        obj.echo("")
        obj.echo("You are running a version of the metaflow service "
                 "that currently doesn't support AWS Step Functions. ")
        obj.echo("For more information on how to upgrade your "
                 "service to a compatible version (>= 2.0.2), visit:")
        obj.echo("    https://admin-docs.metaflow.org/metaflow-on-aws/operation"
                 "s-guide/metaflow-service-migration-guide", fg='green')
        obj.echo("Once you have upgraded your metadata service, please "
                 "re-execute your command.")
        raise IncorrectMetadataServiceVersion("Try again with a more recent "
                                               "version of metaflow service "
                                               "(>=2.0.2).")

def resolve_state_machine_name(obj, name):
    def attach_prefix(name):
      if SFN_STATE_MACHINE_PREFIX is not None:
          return SFN_STATE_MACHINE_PREFIX + '_' + name
      return name
    project = current.get('project_name')
    obj._is_state_machine_name_hashed = False
    if project:
        if name:
            raise MetaflowException("--name is not supported for @projects. "
                                    "Use --branch instead.")
        state_machine_name = attach_prefix(current.project_flow_name)
        project_branch = to_bytes('.'.join((project, current.branch_name)))
        token_prefix = 'mfprj-%s' % to_unicode(
                                        base64.b32encode(
                                            sha1(project_branch).digest()))[:16]
        is_project = True
        # AWS Step Functions has a limit of 80 chars for state machine names.
        # We truncate the state machine name if the computed name is greater
        # than 60 chars and append a hashed suffix to ensure uniqueness.
        if len(state_machine_name) > 60:
            name_hash = to_unicode(
                            base64.b32encode(
                                sha1(to_bytes(state_machine_name)) \
                                  .digest()))[:16].lower()
            state_machine_name =\
                '%s-%s' % (state_machine_name[:60], name_hash)
            obj._is_state_machine_name_hashed = True
    else:
        if name and VALID_NAME.search(name):
            raise MetaflowException(
                "Name '%s' contains invalid characters." % name)

        state_machine_name = attach_prefix(name if name else current.flow_name)
        token_prefix = state_machine_name
        is_project = False

        if len(state_machine_name) > 80:
            msg = "The full name of the workflow:\n*%s*\nis longer than 80 "\
                  "characters.\n\n"\
                  "To deploy this workflow to AWS Step Functions, please "\
                  "assign a shorter name\nusing the option\n"\
                  "*step-functions --name <name> create*." % state_machine_name
            raise StepFunctionsStateMachineNameTooLong(msg)

    return state_machine_name, token_prefix.lower(), is_project

def make_flow(obj,
              token,
              name,
              tags,
              namespace,
              max_workers,
              workflow_timeout,
              is_project):
    datastore = obj.datastore(obj.flow.name,
                              mode='w',
                              metadata=obj.metadata,
                              event_logger=obj.event_logger,
                              monitor=obj.monitor)
    if datastore.TYPE != 's3':
        raise MetaflowException("AWS Step Functions requires --datastore=s3.")

    # Attach AWS Batch decorator to the flow
    decorators._attach_decorators(obj.flow, [BatchDecorator.name])
    decorators._init_step_decorators(
            obj.flow, obj.graph, obj.environment, obj.datastore, obj.logger)

    obj.package = MetaflowPackage(
        obj.flow, obj.environment, obj.echo, obj.package_suffixes)
    package_url = datastore.save_data(
        obj.package.sha, TransformableObject(obj.package.blob))

    return StepFunctions(name,
                         obj.graph,
                         obj.flow,
                         obj.package,
                         package_url,
                         token,
                         obj.metadata,
                         obj.datastore,
                         obj.environment,
                         obj.event_logger,
                         obj.monitor,
                         tags=tags,
                         namespace=namespace,
                         max_workers=max_workers,
                         username=get_username(),
                         workflow_timeout=workflow_timeout,
                         is_project=is_project)

def resolve_token(name,
                  token_prefix,
                  obj,
                  authorize,
                  given_token,
                  generate_new_token,
                  is_project):

    # 1) retrieve the previous deployment, if one exists
    workflow = StepFunctions.get_existing_deployment(name)
    if workflow is None:
        obj.echo("It seems this is the first time you are deploying *%s* to "
                 "AWS Step Functions." % name)
        prev_token = None
    else:
        prev_user, prev_token = workflow

    # 2) authorize this deployment
    if prev_token is not None:
        if authorize is None:
            authorize = load_token(token_prefix)
        elif authorize.startswith('production:'):
            authorize = authorize[11:]

        # we allow the user who deployed the previous version to re-deploy,
        # even if they don't have the token
        if prev_user != get_username() and authorize != prev_token:
            obj.echo("There is an existing version of *%s* on AWS Step "
                     "Functions which was deployed by the user "
                     "*%s*." % (name, prev_user))
            obj.echo("To deploy a new version of this flow, you need to use "
                     "the same production token that they used. ")
            obj.echo("Please reach out to them to get the token. Once you "
                     "have it, call this command:")
            obj.echo("    step-functions create --authorize MY_TOKEN", 
                        fg='green')
            obj.echo('See "Organizing Results" at docs.metaflow.org for more '
                     "information about production tokens.")
            raise IncorrectProductionToken("Try again with the correct "
                                           "production token.")

    # 3) do we need a new token or should we use the existing token?
    if given_token:
        if is_project:
            # we rely on a known prefix for @project tokens, so we can't
            # allow the user to specify a custom token with an arbitrary prefix
            raise MetaflowException("--new-token is not supported for "
                                    "@projects. Use --generate-new-token to "
                                    "create a new token.")
        if given_token.startswith('production:'):
            given_token = given_token[11:]
        token = given_token
        obj.echo('')
        obj.echo("Using the given token, *%s*." % token)
    elif prev_token is None or generate_new_token:
        token = new_token(token_prefix, prev_token)
        if token is None:
            if prev_token is None:
                raise MetaflowInternalError("We could not generate a new "
                                            "token. This is unexpected. ")
            else:
                raise MetaflowException("--generate-new-token option is not "
                                        "supported after using --new-token. "
                                        "Use --new-token to make a new "
                                        "namespace.")
        obj.echo('')
        obj.echo("A new production token generated.")
    else:
        token = prev_token

    obj.echo('')
    obj.echo("The namespace of this production flow is")
    obj.echo('    production:%s' % token, fg='green')
    obj.echo("To analyze results of this production flow "
             "add this line in your notebooks:")
    obj.echo('    namespace(\"production:%s\")' % token, fg='green')
    obj.echo("If you want to authorize other people to deploy new versions "
             "of this flow to AWS Step Functions, they need to call")
    obj.echo("    step-functions create --authorize %s" % token, fg='green')
    obj.echo("when deploying this flow to AWS Step Functions for the first "
             "time.")
    obj.echo('See "Organizing Results" at https://docs.metaflow.org/ for more '
             "information about production tokens.")
    obj.echo('')
    store_token(token_prefix, token)
    return token


@parameters.add_custom_parameters(deploy_mode=False)
@step_functions.command(help="Trigger the workflow on AWS Step Functions.")
@click.option('--run-id-file',
                default=None,
                show_default=True,
                type=str,
                help="Write the ID of this run to the file specified.")
@click.pass_obj
def trigger(obj, run_id_file=None, **kwargs):
    def _convert_value(param):
        # Swap `-` with `_` in parameter name to match click's behavior
        val = kwargs.get(param.name.replace('-', '_').lower())
        return json.dumps(val) if param.kwargs.get('type') == JSONType else \
            val() if callable(val) else val

    params = {param.name: _convert_value(param)
              for _, param in obj.flow._get_parameters()
                if kwargs.get(param.name.replace('-', '_').lower()) is not None}

    response = StepFunctions.trigger(obj.state_machine_name, params)

    id = response['executionArn'].split(':')[-1]
    run_id = 'sfn-' + id

    if run_id_file:
        with open(run_id_file, 'w') as f:
            f.write(str(run_id))

    obj.echo("Workflow *{name}* triggered on AWS Step Functions "
        "(run-id *{run_id}*)."
            .format(name=obj.state_machine_name, run_id=run_id), bold=True)


@step_functions.command(
    help="List all runs of the workflow on AWS Step Functions.")
@click.option("--running", default=False, is_flag=True,
    help="List all runs of the workflow in RUNNING state on "
        "AWS Step Functions.")
@click.option("--succeeded", default=False, is_flag=True,
    help="List all runs of the workflow in SUCCEEDED state on "
        "AWS Step Functions.")
@click.option("--failed", default=False, is_flag=True,
    help="List all runs of the workflow in FAILED state on "
        "AWS Step Functions.")
@click.option("--timed-out", default=False, is_flag=True,
    help="List all runs of the workflow in TIMED_OUT state on "
        "AWS Step Functions.")
@click.option("--aborted", default=False, is_flag=True,
    help="List all runs of the workflow in ABORTED state on "
        "AWS Step Functions.")
@click.pass_obj
def list_runs(obj, 
              running=False,
              succeeded=False,
              failed=False,
              timed_out=False,
              aborted=False):
    states = []
    if running:
        states.append('RUNNING')
    if succeeded:
        states.append('SUCCEEDED')
    if failed:
        states.append('FAILED')
    if timed_out:
        states.append('TIMED_OUT')
    if aborted:
        states.append('ABORTED')
    executions = StepFunctions.list(obj.state_machine_name, states)
    found = False
    for execution in executions:
        found = True
        if execution.get('stopDate'):
            obj.echo(
                "*sfn-{id}* "
                "startedAt:'{startDate}' "
                "stoppedAt:'{stopDate}' "
                "*{status}*".format(
                    id=execution['name'], 
                    status=execution['status'], 
                    startDate=execution['startDate'].replace(microsecond=0), 
                    stopDate=execution['stopDate'].replace(microsecond=0),
                )
            )
        else:
            obj.echo(
                "*sfn-{id}* "
                "startedAt:'{startDate}' "
                "*{status}*".format(
                    id=execution['name'], 
                    status=execution['status'], 
                    startDate=execution['startDate'].replace(microsecond=0)
                )
            )
    if not found:
        if len(states) > 0:
            status = ''
            for idx, state in enumerate(states):
                if idx == 0:
                    pass
                elif idx == len(states) - 1:
                    status += ' and '
                else:
                    status += ', '
                status += '*%s*' % state
            obj.echo('No %s executions for *%s* found on AWS Step Functions.'
                % (status, obj.state_machine_name))
        else:
            obj.echo('No executions for *%s* found on AWS Step Functions.' \
                % (obj.state_machine_name))



class Override(object):
    def __init__(self, obj_mapping, wrapped_function):
        self._obj_mapping = obj_mapping
        self._wrapped = wrapped_function

    @property
    def obj_mapping(self):
        return self._obj_mapping

    @property
    def func(self):
        return self._wrapped


class AttrOverride(Override):
    def __init__(self, is_setattr, obj_mapping, wrapped_function):
        super(AttrOverride, self).__init__(obj_mapping, wrapped_function)
        self._is_setattr = is_setattr

    @property
    def is_setattr(self):
        return self._is_setattr


class LocalOverride(Override):
    pass


class LocalAttrOverride(AttrOverride):
    pass


class RemoteOverride(Override):
    pass


class RemoteAttrOverride(AttrOverride):
    pass


def local_override(obj_mapping):
    if not isinstance(obj_mapping, dict):
        raise ValueError(
            "@local_override takes a dictionary: <class name> -> [<overriden method>]"
        )

    def _wrapped(func):
        return LocalOverride(obj_mapping, func)

    return _wrapped


def local_getattr_override(obj_mapping):
    if not isinstance(obj_mapping, dict):
        raise ValueError(
            "@local_getattr_override takes a dictionary: <class name> -> [<overriden attribute>]"
        )

    def _wrapped(func):
        return LocalAttrOverride(False, obj_mapping, func)

    return _wrapped


def local_setattr_override(obj_mapping):
    if not isinstance(obj_mapping, dict):
        raise ValueError(
            "@local_setattr_override takes a dictionary: <class name> -> [<overriden attribute>]"
        )

    def _wrapped(func):
        return LocalAttrOverride(True, obj_mapping, func)

    return _wrapped


def remote_override(obj_mapping):
    if not isinstance(obj_mapping, dict):
        raise ValueError(
            "@remote_override takes a dictionary: <class name> -> [<overriden method>]"
        )

    def _wrapped(func):
        return RemoteOverride(obj_mapping, func)

    return _wrapped


def remote_getattr_override(obj_mapping):
    if not isinstance(obj_mapping, dict):
        raise ValueError(
            "@remote_getattr_override takes a dictionary: <class name> -> [<overriden attribute>]"
        )

    def _wrapped(func):
        return RemoteAttrOverride(False, obj_mapping, func)

    return _wrapped


def remote_setattr_override(obj_mapping):
    if not isinstance(obj_mapping, dict):
        raise ValueError(
            "@remote_setattr_override takes a dictionary: <class name> -> [<overriden attribute>]"
        )

    def _wrapped(func):
        return RemoteAttrOverride(True, obj_mapping, func)

    return _wrapped


class LocalException(object):
    def __init__(self, class_path, wrapped_class):
        self._class_path = class_path
        self._class = wrapped_class

    @property
    def class_path(self):
        return self._class_path

    @property
    def wrapped_class(self):
        return self._class


class RemoteExceptionSerializer(object):
    def __init__(self, class_path, serializer):
        self._class_path = class_path
        self._serializer = serializer

    @property
    def class_path(self):
        return self._class_path

    @property
    def serializer(self):
        return self._serializer


def local_exception(class_path):
    def _wrapped(cls):
        return LocalException(class_path, cls)

    return _wrapped


def remote_exception_serialize(class_path):
    def _wrapped(func):
        return RemoteExceptionSerializer(class_path, func)

    return _wrapped



import click
import json
import os
import shutil

from os.path import expanduser

from metaflow.datastore.local import LocalDataStore
from metaflow.metaflow_config import DATASTORE_LOCAL_DIR, DEFAULT_METADATA
from metaflow.util import to_unicode


def makedirs(path):
    # This is for python2 compatibility.
    # Python3 has os.makedirs(exist_ok=True).
    try:
        os.makedirs(path)
    except OSError as x:
        if x.errno == 17:
            return
        else:
            raise

def echo_dev_null(*args, **kwargs):
    pass

def echo_always(line, **kwargs):
    click.secho(line, **kwargs)

@click.group(invoke_without_command=True)
@click.pass_context
def main(ctx):
    global echo
    echo = echo_always

    import metaflow
    echo('Metaflow ',
         fg='magenta',
         bold=True,
         nl=False)

    if ctx.invoked_subcommand is None:
        echo('(%s): ' % metaflow.__version__,
             fg='magenta',
             bold=False,
             nl=False)
    else:
        echo('(%s)\n' % metaflow.__version__,
             fg='magenta',
             bold=False)

    if ctx.invoked_subcommand is None:
        echo("More data science, less engineering\n",
             fg='magenta')

        # metaflow URL
        echo('http://docs.metaflow.org', fg='cyan', nl=False)
        echo(' - Read the documentation')

        # metaflow chat
        echo('http://chat.metaflow.org', fg='cyan', nl=False)
        echo(' - Chat with us')

        # metaflow help email
        echo('help@metaflow.org', fg='cyan', nl=False)
        echo('        - Get help by email\n')

        # print a short list of next steps.
        short_help = {'tutorials': 'Browse and access metaflow tutorials.',
                      'configure': 'Configure metaflow to access the cloud.',
                      'status': 'Display the current working tree.',
                      'help': 'Show all available commands to run.'}

        echo('Commands:', bold=False)

        for cmd, desc in short_help.items():
            echo('  metaflow {0:<10} '.format(cmd),
                 fg='cyan',
                 bold=False,
                 nl=False)

            echo('%s' % desc)

@main.command(help='Show all available commands.')
@click.pass_context
def help(ctx):
    print(ctx.parent.get_help())

@main.command(help='Show flows accessible from the current working tree.')
def status():
    from metaflow.client import get_metadata
    res = get_metadata()
    if res:
        res = res.split('@')
    else:
        raise click.ClickException('Unknown status: cannot find a Metadata provider')
    if res[0] == 'service':
        echo('Using Metadata provider at: ', nl=False)
        echo('"%s"\n' % res[1], fg='cyan')
        echo('To list available flows, type:\n')
        echo('1. python')
        echo('2. from metaflow import Metaflow')
        echo('3. list(Metaflow())')
        return

    from metaflow.client import namespace, metadata, Metaflow

    # Get the local data store path
    path = LocalDataStore.get_datastore_root_from_config(echo,
                                                         create_on_absent=False)
    # Throw an exception
    if path is None:
        raise click.ClickException("Could not find " +\
                                   click.style('"%s"' % DATASTORE_LOCAL_DIR,
                                               fg='red') +\
                                   " in the current working tree.")

    stripped_path = os.path.dirname(path)
    namespace(None)
    metadata('local@%s' % stripped_path)
    echo('Working tree found at: ', nl=False)
    echo('"%s"\n' % stripped_path, fg='cyan')
    echo('Available flows:', fg='cyan', bold=True)
    for flow in Metaflow():
        echo('* %s' % flow, fg='cyan')

@main.group(help="Browse and access the metaflow tutorial episodes.")
def tutorials():
    pass

def get_tutorials_dir():
    metaflow_dir = os.path.dirname(__file__)
    package_dir = os.path.dirname(metaflow_dir)
    tutorials_dir = os.path.join(package_dir, 'metaflow', 'tutorials')

    return tutorials_dir

def get_tutorial_metadata(tutorial_path):
    metadata = {}
    with open(os.path.join(tutorial_path, 'README.md')) as readme:
        content =  readme.read()

    paragraphs = [paragraph.strip() \
                  for paragraph \
                  in content.split('#') if paragraph]
    metadata['description'] = paragraphs[0].split('**')[1]
    header = paragraphs[0].split('\n')
    header = header[0].split(':')
    metadata['episode'] = header[0].strip()[len('Episode '):]
    metadata['title'] = header[1].strip()

    for paragraph in paragraphs[1:]:
        if paragraph.startswith('Before playing'):
            lines = '\n'.join(paragraph.split('\n')[1:])
            metadata['prereq'] = lines.replace('```', '')

        if paragraph.startswith('Showcasing'):
            lines = '\n'.join(paragraph.split('\n')[1:])
            metadata['showcase'] = lines.replace('```', '')

        if paragraph.startswith('To play'):
            lines = '\n'.join(paragraph.split('\n')[1:])
            metadata['play'] = lines.replace('```', '')

    return metadata

def get_all_episodes():
    episodes = []
    for name in sorted(os.listdir(get_tutorials_dir())):
        # Skip hidden files (like .gitignore)
        if not name.startswith('.'):
            episodes.append(name)
    return episodes

@tutorials.command(help="List the available episodes.")
def list():
    echo('Episodes:', fg='cyan', bold=True)
    for name in get_all_episodes():
        path = os.path.join(get_tutorials_dir(), name)
        metadata = get_tutorial_metadata(path)
        echo('* {0: <20} '.format(metadata['episode']),
             fg='cyan',
             nl=False)
        echo('- {0}'.format(metadata['title']))

    echo('\nTo pull the episodes, type: ')
    echo('metaflow tutorials pull', fg='cyan')

def validate_episode(episode):
    src_dir = os.path.join(get_tutorials_dir(), episode)
    if not os.path.isdir(src_dir):
        raise click.BadArgumentUsage("Episode " + \
                                     click.style("\"{0}\"".format(episode),
                                                 fg='red') + " does not exist."\
                                     " To see a list of available episodes, "\
                                     "type:\n" + \
                                     click.style("metaflow tutorials list",
                                                 fg='cyan'))

def autocomplete_episodes(ctx, args, incomplete):
    return [k for k in get_all_episodes() if incomplete in k]

@tutorials.command(help="Pull episodes "\
                   "into your current working directory.")
@click.option('--episode', default="", help="Optional episode name "\
              "to pull only a single episode.")
def pull(episode):
    tutorials_dir = get_tutorials_dir()
    if not episode:
        episodes = get_all_episodes()
    else:
        episodes = [episode]
        # Validate that the list is valid.
        for episode in episodes:
            validate_episode(episode)
    # Create destination `metaflow-tutorials` dir.
    dst_parent = os.path.join(os.getcwd(), 'metaflow-tutorials')
    makedirs(dst_parent)

    # Pull specified episodes.
    for episode in episodes:
        dst_dir = os.path.join(dst_parent, episode)
        # Check if episode has already been pulled before.
        if os.path.exists(dst_dir):
            if click.confirm("Episode " + \
                             click.style("\"{0}\"".format(episode), fg='red') +\
                             " has already been pulled before. Do you wish "\
                             "to delete the existing version?"):
                shutil.rmtree(dst_dir)
            else:
                continue
        echo('Pulling episode ', nl=False)
        echo('\"{0}\"'.format(episode), fg='cyan', nl=False)
        # TODO: Is the following redudant?
        echo(' into your current working directory.')
        # Copy from (local) metaflow package dir to current.
        src_dir = os.path.join(tutorials_dir, episode)
        shutil.copytree(src_dir, dst_dir)

    echo('\nTo know more about an episode, type:\n', nl=False)
    echo('metaflow tutorials info [EPISODE]', fg='cyan')

@tutorials.command(help='Find out more about an episode.')
@click.argument('episode', autocompletion=autocomplete_episodes)
def info(episode):
    validate_episode(episode)
    src_dir = os.path.join(get_tutorials_dir(), episode)
    metadata = get_tutorial_metadata(src_dir)
    echo('Synopsis:', fg='cyan', bold=True)
    echo('%s' % metadata['description'])

    echo('\nShowcasing:', fg='cyan', bold=True, nl=True)
    echo('%s' % metadata['showcase'])

    if 'prereq' in metadata:
        echo('\nBefore playing:', fg='cyan', bold=True, nl=True)
        echo('%s' % metadata['prereq'])

    echo('\nTo play:', fg='cyan', bold=True)
    echo('%s' % metadata['play'])

# NOTE: This code needs to be in sync with metaflow/metaflow_config.py.
METAFLOW_CONFIGURATION_DIR =\
    expanduser(os.environ.get('METAFLOW_HOME', '~/.metaflowconfig'))

@main.group(help="Configure Metaflow to access the cloud.")
def configure():
    makedirs(METAFLOW_CONFIGURATION_DIR)

def get_config_path(profile):
    config_file = 'config.json' if not profile else ('config_%s.json' % profile)
    path = os.path.join(METAFLOW_CONFIGURATION_DIR, config_file)
    return path

def overwrite_config(profile):
    path = get_config_path(profile)
    if os.path.exists(path):
        if not click.confirm(
            click.style('We found an existing configuration for your ' +
                        'profile. Do you want to modify the existing ' +
                        'configuration?', fg='red', bold=True)):
            echo('You can configure a different named profile by using the '
                 '--profile argument. You can activate this profile by setting '
                 'the environment variable METAFLOW_PROFILE to the named '
                 'profile.', fg='yellow')
            return False
    return True


def check_for_missing_profile(profile):
    path = get_config_path(profile)
    # Absence of default config is equivalent to running locally.
    if profile and not os.path.exists(path):
        raise click.ClickException("Couldn't find configuration for profile " +
                                   click.style('"%s"' % profile, fg='red') +
                                   " in " +
                                   click.style('"%s"' % path, fg='red'))

def get_env(profile):
    path = get_config_path(profile)
    if os.path.exists(path):
        with open(path) as f:
            return json.load(f)
    return {}

def persist_env(env_dict, profile):
    # TODO: Should we persist empty env_dict or notify user differently?
    path = get_config_path(profile)

    with open(path, 'w') as f:
        json.dump(env_dict, f, indent=4, sort_keys=True)

    echo('\nConfiguration successfully written to ', nl=False, bold=True)
    echo('"%s"' % path, fg='cyan')

@configure.command(help='Reset configuration to disable cloud access.')
@click.option('--profile', '-p', default='',
              help="Optional named profile.")
def reset(profile):
    check_for_missing_profile(profile)
    path = get_config_path(profile)
    if os.path.exists(path):
        if click.confirm('Do you really wish to reset the configuration in ' +\
                         click.style('"%s"' % path, fg='cyan'), abort=True):
            os.remove(path)
            echo('Configuration successfully reset to run locally.')
    else:
        echo('Configuration is already reset to run locally.')

@configure.command(help='Show existing configuration.')
@click.option('--profile', '-p', default='',
              help="Optional named profile.")
def show(profile):
    check_for_missing_profile(profile)
    path = get_config_path(profile)
    env_dict = {}
    if os.path.exists(path):
        with open(path, 'r') as f:
            env_dict = json.load(f)
    if env_dict:
        echo('Showing configuration in ', nl=False)
        echo('"%s"\n' % path, fg='cyan')
        for k,v in env_dict.items():
            echo('%s=%s' % (k, v))
    else:
        echo('Configuration is set to run locally.')

@configure.command(help='Export configuration to a file.')
@click.option('--profile', '-p', default='',
              help="Optional named profile whose configuration must be "
                   "exported.")
@click.argument('output_filename', type=click.Path(resolve_path=True))
def export(profile, output_filename):
    check_for_missing_profile(profile)
    # Export its contents to a new file.
    path = get_config_path(profile)
    env_dict = {}
    if os.path.exists(path):
        with open(path, 'r') as f:
            env_dict = json.load(f)
    # resolve_path doesn't expand `~` in `path`.
    output_path = expanduser(output_filename)
    if os.path.exists(output_path):
        if click.confirm('Do you wish to overwrite the contents in ' +
                         click.style('"%s"' % output_path, fg='cyan') + '?',
                         abort=True):
            pass
    # Write to file.
    with open(output_path, 'w') as f:
        json.dump(env_dict, f, indent=4, sort_keys=True)
    echo('Configuration successfully exported to: ', nl=False)
    echo('"%s"' % output_path, fg='cyan')

@configure.command(help='Import configuration from a file.', name='import')
@click.option('--profile', '-p', default='',
              help="Optional named profile to which the configuration must be "
                   "imported into.")
@click.argument('input_filename', type=click.Path(exists=True,
                                                  resolve_path=True))
def import_from(profile, input_filename):
    check_for_missing_profile(profile)
    # Import configuration.
    input_path = expanduser(input_filename)
    env_dict = {}
    with open(input_path, 'r') as f:
        env_dict = json.load(f)
    echo('Configuration successfully read from: ', nl=False)
    echo('"%s"' % input_path, fg='cyan')

    # Persist configuration.
    overwrite_config(profile)
    persist_env(env_dict, profile)

@configure.command(help='Configure metaflow to access hosted sandbox.')
@click.option('--profile', '-p', default='',
                help='Configure a named profile. Activate the profile by setting ' 
                    '`METAFLOW_PROFILE` environment variable.')
def sandbox(profile):
    overwrite_config(profile)
    # Prompt for user input.
    encoded_str = click.prompt('Following instructions from '
                               'https://metaflow.org/sandbox, '
                               'please paste the encoded magic string',
                               type=str)
    # Decode the bytes to env_dict.
    try:
        import base64, zlib
        from metaflow.util import to_bytes
        env_dict =\
            json.loads(to_unicode(zlib.decompress(base64.b64decode(to_bytes(encoded_str)))))
    except:
        # TODO: Add the URL for contact us page in the error?
        raise click.BadArgumentUsage('Could not decode the sandbox '\
                                     'configuration. Please contact us.')
    # Persist to a file.
    persist_env(env_dict, profile)

@configure.command(help='Configure metaflow to access self-managed AWS resources.')
@click.option('--profile', '-p', default='',
                help='Configure a named profile. Activate the profile by setting ' 
                    '`METAFLOW_PROFILE` environment variable.')
@click.pass_context
def aws(ctx, profile):
    def cyan(string):
        return click.style(string, fg='cyan')

    def yellow(string):
        return click.style(string, fg='yellow')

    # Greet the user!
    echo('Welcome to Metaflow! Follow the prompts to configure your '
         'installation.\n',
         bold=True)

    # Check for existing configuration.
    if not overwrite_config(profile):
        ctx.abort()

    # Verify that the user has configured AWS credentials on their computer.
    if not click.confirm('\nMetaflow relies on ' +  
                         yellow('AWS access credentials') + 
                         ' present on your computer to access resources on AWS.'
                         '\nBefore proceeding further, please confirm that you '
                         'have already configured these access credentials on '
                         'this computer.',
                         default=True):
        echo('There are many ways to setup your AWS access credentials. You '
             'can get started by following this guide: ',
             nl=False,
             fg='yellow')
        echo('https://docs.aws.amazon.com/cli/latest/userguide/cli-chap-configure.html',
             fg='cyan')
        ctx.abort()

    existing_env = get_env(profile)
    empty_profile = False
    if not existing_env:
        empty_profile = True
    env = {}

    # Configure Amazon S3 as the datastore.
    use_s3_as_datastore = click.confirm('\nMetaflow can use ' +
                            yellow('Amazon S3 as the storage backend') +
                            ' for all code and data artifacts on ' +
                            'AWS.\nAmazon S3 is a strict requirement if you ' +
                            'intend to execute your flows on AWS Batch ' +
                            'and/or schedule them on AWS Step ' +
                            'Functions.\nWould you like to configure Amazon ' +
                            'S3 as the default storage backend?',
                            default=empty_profile or \
                            existing_env.get(
                                'METAFLOW_DEFAULT_DATASTORE', '') == 's3',
                            abort=False)
    if use_s3_as_datastore:
        # Set Amazon S3 as default datastore.
        env['METAFLOW_DEFAULT_DATASTORE'] = 's3'
        # Set Amazon S3 folder for datastore.
        env['METAFLOW_DATASTORE_SYSROOT_S3'] =\
            click.prompt(cyan('[METAFLOW_DATASTORE_SYSROOT_S3]') + 
                         ' Amazon S3 folder for Metaflow artifact storage ' +
                         '(s3://<bucket>/<prefix>).',
                         default=\
                            existing_env.get('METAFLOW_DATASTORE_SYSROOT_S3'),
                         show_default=True)
        # Set Amazon S3 folder for datatools.
        env['METAFLOW_DATATOOLS_SYSROOT_S3'] =\
                click.prompt(cyan('[METAFLOW_DATATOOLS_SYSROOT_S3]') + 
                             yellow(' (optional)') + 
                             ' Amazon S3 folder for Metaflow datatools ' +
                             '(s3://<bucket>/<prefix>).',
                             default=\
                                existing_env.get('METAFLOW_DATATOOLS_SYSROOT_S3',
                                    os.path.join(
                                        env['METAFLOW_DATASTORE_SYSROOT_S3'],
                                            'data')),
                             show_default=True)

    # Configure Metadata service for tracking.
    if click.confirm('\nMetaflow can use a ' +
                     yellow('remote Metadata Service to track') + 
                     ' and persist flow execution metadata.\nConfiguring the '
                     'service is a requirement if you intend to schedule your '
                     'flows with AWS Step Functions.\nWould you like to '
                     'configure the Metadata Service?',
                     default=empty_profile or\
                        existing_env.get('METAFLOW_DEFAULT_METADATA', '') ==\
                            'service' or\
                        'METAFLOW_SFN_IAM_ROLE' in env,
                     abort=False):
        # Set Metadata Service as default.
        env['METAFLOW_DEFAULT_METADATA'] = 'service'
        # Set URL for the Metadata Service.
        env['METAFLOW_SERVICE_URL'] =\
                click.prompt(cyan('[METAFLOW_SERVICE_URL]') +
                             ' URL for Metaflow Service.',
                             default=existing_env.get('METAFLOW_SERVICE_URL'),
                             show_default=True)
        # Set internal URL for the Metadata Service.
        env['METAFLOW_SERVICE_INTERNAL_URL'] =\
                click.prompt(cyan('[METAFLOW_SERVICE_INTERNAL_URL]') +
                             yellow(' (optional)') +
                             ' URL for Metaflow Service ' +
                             '(Accessible only within VPC).',
                             default=\
                                existing_env.get('METAFLOW_SERVICE_INTERNAL_URL',
                                    env['METAFLOW_SERVICE_URL']),
                             show_default=True)
        # Set Auth Key for the Metadata Service.
        env['METAFLOW_SERVICE_AUTH_KEY'] =\
                click.prompt(cyan('[METAFLOW_SERVICE_AUTH_KEY]') + 
                             yellow(' (optional)') +
                             ' Auth Key for Metaflow Service.',
                             default=\
                                existing_env.get('METAFLOW_SERVICE_AUTH_KEY', ''),
                             show_default=True)

    # Configure AWS Batch for compute.
    if use_s3_as_datastore:
        if click.confirm('\nMetaflow can scale your flows by ' +
                         yellow('executing your steps on AWS Batch') + 
                         '.\nAWS Batch is a strict requirement if you intend '
                         'to schedule your flows on AWS Step Functions.\nWould '
                         'you like to configure AWS Batch as your compute '
                         'backend?',
                         default=empty_profile or 
                            'METAFLOW_BATCH_JOB_QUEUE' in existing_env,
                         abort=False):
            # Set AWS Batch Job Queue.
            env['METAFLOW_BATCH_JOB_QUEUE'] =\
                    click.prompt(cyan('[METAFLOW_BATCH_JOB_QUEUE]') +
                                 ' AWS Batch Job Queue.',
                                 default=\
                                    existing_env.get('METAFLOW_BATCH_JOB_QUEUE'),
                                 show_default=True)
            # Set IAM role for AWS Batch jobs to assume.
            env['METAFLOW_ECS_S3_ACCESS_IAM_ROLE'] =\
                    click.prompt(cyan('[METAFLOW_ECS_S3_ACCESS_IAM_ROLE]') + 
                                 ' IAM role for AWS Batch jobs to access AWS ' +
                                 'resources (Amazon S3 etc.).',
                                 default=\
                                    existing_env.get('METAFLOW_ECS_S3_ACCESS_IAM_ROLE'),
                                 show_default=True)
            # Set default Docker repository for AWS Batch jobs.
            env['METAFLOW_BATCH_CONTAINER_REGISTRY'] =\
                    click.prompt(cyan('[METAFLOW_BATCH_CONTAINER_REGISTRY]') + 
                                 yellow(' (optional)') +
                                 ' Default Docker image repository for AWS ' +
                                 'Batch jobs. If nothing is specified, ' +
                                 'dockerhub (hub.docker.com/) is ' +
                                 'used as default.',
                                 default=\
                                    existing_env.get('METAFLOW_BATCH_CONTAINER_REGISTRY', ''),
                                 show_default=True)
            # Set default Docker image for AWS Batch jobs.
            env['METAFLOW_BATCH_CONTAINER_IMAGE'] =\
                    click.prompt(cyan('[METAFLOW_BATCH_CONTAINER_IMAGE]') + 
                                 yellow(' (optional)') +
                                 ' Default Docker image for AWS Batch jobs. ' +
                                 'If nothing is specified, an appropriate ' +
                                 'python image is used as default.',
                                 default=\
                                    existing_env.get('METAFLOW_BATCH_CONTAINER_IMAGE', ''),
                                 show_default=True)

            # Configure AWS Step Functions for scheduling.
            if click.confirm('\nMetaflow can ' +
                             yellow('schedule your flows on AWS Step '
                                    'Functions') + 
                             ' and trigger them at a specific cadence using '
                             'Amazon EventBridge.\nTo support flows involving '
                             'foreach steps, you would need access to AWS '
                             'DynamoDB.\nWould you like to configure AWS Step '
                             'Functions for scheduling?',
                             default=empty_profile or
                                'METAFLOW_SFN_IAM_ROLE' in existing_env,
                             abort=False):
                # Configure IAM role for AWS Step Functions.
                env['METAFLOW_SFN_IAM_ROLE'] =\
                        click.prompt(cyan('[METAFLOW_SFN_IAM_ROLE]') + 
                                     ' IAM role for AWS Step Functions to ' +
                                     'access AWS resources (AWS Batch, ' +
                                     'AWS DynamoDB).',
                                     default=\
                                        existing_env.get('METAFLOW_SFN_IAM_ROLE'),
                                     show_default=True)
                # Configure IAM role for AWS Events Bridge.
                env['METAFLOW_EVENTS_SFN_ACCESS_IAM_ROLE'] =\
                        click.prompt(cyan('[METAFLOW_EVENTS_SFN_ACCESS_IAM_ROLE]') + 
                                     ' IAM role for Amazon EventBridge to ' +
                                     'access AWS Step Functions.',
                                     default=\
                                        existing_env.get('METAFLOW_EVENTS_SFN_ACCESS_IAM_ROLE'),
                                     show_default=True)
                # Configure AWS DynamoDB Table for AWS Step Functions.
                env['METAFLOW_SFN_DYNAMO_DB_TABLE'] =\
                        click.prompt(cyan('[METAFLOW_SFN_DYNAMO_DB_TABLE]') + 
                                    ' AWS DynamoDB table name for tracking '+
                                    'AWS Step Functions execution metadata.',
                                     default=\
                                        existing_env.get('METAFLOW_SFN_DYNAMO_DB_TABLE'),
                                     show_default=True)
    persist_env({k: v for k, v in env.items() if v}, profile)



import json

# Define message enums
# Unfortunately we can't use enums because they are not supported
# officially in Python2
class MessageTypes(object):
    SHUTDOWN, LOG_EVENT = range(1, 3)

class Message(object):

    def __init__(self, msg_type, payload):
        self.msg_type = msg_type
        self.payload = payload

    def serialize(self):
        msg = {
            'msg_type': self.msg_type,
            'payload': self.payload,
        }
        return json.dumps(msg)+"\n"

def deserialize(json_msg):
    return Message(**json.loads(json_msg))


import os
import sys
import tarfile
import json
from hashlib import sha1
from io import BytesIO
from itertools import chain

from .metaflow_config import DEFAULT_PACKAGE_SUFFIXES
from .util import to_unicode
from . import R

DEFAULT_SUFFIXES_LIST = DEFAULT_PACKAGE_SUFFIXES.split(',')


class MetaflowPackage(object):

    def __init__(self, flow, environment, echo, suffixes=DEFAULT_SUFFIXES_LIST):
        self.suffixes = list(set().union(suffixes, DEFAULT_SUFFIXES_LIST))
        self.environment = environment
        self.metaflow_root = os.path.dirname(__file__)
        try:
            import metaflow_custom
        except ImportError:
            self.metaflow_custom_root = None
        else:
            self.metaflow_custom_root = os.path.dirname(metaflow_custom.__file__)
            self.metaflow_custom_addl_suffixes = getattr(
                metaflow_custom,
                'METAFLOW_CUSTOM_PACKAGE_SUFFIXES',
                None)

        environment.init_environment(echo)
        for step in flow:
            for deco in step.decorators:
                deco.package_init(flow,
                                  step.__name__,
                                  environment)
        self.blob, self.sha = self._make()

    def _walk(self, root, exclude_hidden=True, addl_suffixes=None):
        if addl_suffixes is None:
            addl_suffixes = []
        root = to_unicode(root)  # handle files/folder with non ascii chars
        prefixlen = len('%s/' % os.path.dirname(root))
        for path, dirs, files in os.walk(root):
            if exclude_hidden and '/.' in path:
                continue
            # path = path[2:] # strip the ./ prefix
            # if path and (path[0] == '.' or './' in path):
            #    continue
            for fname in files:
                if fname[0] == '.':
                    continue
                if any(fname.endswith(suffix) for suffix in self.suffixes + addl_suffixes):
                    p = os.path.join(path, fname)
                    yield p, p[prefixlen:]

    def path_tuples(self):
        """
        Returns list of (path, arcname) to be added to the job package, where
        `arcname` is the alternative name for the file in the package.
        """
        # We want the following contents in the tarball
        # Metaflow package itself
        for path_tuple in self._walk(self.metaflow_root, exclude_hidden=False):
            yield path_tuple
        # Metaflow customization if any
        if self.metaflow_custom_root:
            for path_tuple in self._walk(
                    self.metaflow_custom_root,
                    exclude_hidden=False,
                    addl_suffixes=self.metaflow_custom_addl_suffixes):
                yield path_tuple
        # the package folders for environment
        for path_tuple in self.environment.add_to_package():
            yield path_tuple
        if R.use_r():
            # the R working directory
            for path_tuple in self._walk('%s/' % R.working_dir()):
                yield path_tuple
            # the R package
            for path_tuple in R.package_paths():
                yield path_tuple
        else:
            # the user's working directory
            flowdir = os.path.dirname(os.path.abspath(sys.argv[0])) + '/'
            for path_tuple in self._walk(flowdir):
                yield path_tuple

    def _add_info(self, tar):
        info = tarfile.TarInfo('INFO')
        env = self.environment.get_environment_info()
        buf = BytesIO()
        buf.write(json.dumps(env).encode('utf-8'))
        buf.seek(0)
        info.size = len(buf.getvalue())
        tar.addfile(info, buf)

    def _make(self):
        def no_mtime(tarinfo):
            # a modification time change should not change the hash of
            # the package. Only content modifications will.
            tarinfo.mtime = 0
            return tarinfo

        buf = BytesIO()
        with tarfile.TarFile(fileobj=buf, mode='w') as tar:
            self._add_info(tar)
            for path, arcname in self.path_tuples():
                tar.add(path, arcname=arcname,
                        recursive=False, filter=no_mtime)

        blob = buf.getvalue()
        return blob, sha1(blob).hexdigest()

    def __str__(self):
        return '<code package %s>' % self.sha



import sys


class DebugEventLogger(object):
    TYPE = 'debugLogger'

    def log(self, msg):
        sys.stderr.write('event_logger: ' + str(msg)+'\n')

    def process_message(self, msg):
        # type: (Message) -> None
        self.log(msg.payload)

    def shutdown(self):
        pass



import sys
import types

_expected_extensions = {
    'FLOW_DECORATORS': [],
    'STEP_DECORATORS': [],
    'ENVIRONMENTS': [],
    'METADATA_PROVIDERS': [],
    'SIDECARS': {},
    'LOGGING_SIDECARS': {},
    'MONITOR_SIDECARS': {},
    'AWS_CLIENT_PROVIDERS': [],
    'get_plugin_cli': lambda : []
}

try:
    import metaflow_custom.plugins as _ext_plugins
except ImportError as e:
    ver = sys.version_info[0] * 10 + sys.version_info[1]
    if ver >= 36:
        # e.name is set to the name of the package that fails to load
        # so don't error ONLY IF the error is importing this module (but do
        # error if there is a transitive import error)
        if not (isinstance(e, ModuleNotFoundError) and \
                e.name in ['metaflow_custom', 'metaflow_custom.plugins']):
            print(
                "Cannot load metaflow_custom plugins -- "
                "if you want to ignore, uninstall metaflow_custom package")
            raise
    class _fake(object):
        def __getattr__(self, name):
            if name in _expected_extensions:
                return _expected_extensions[name]
            raise AttributeError

    _ext_plugins = _fake()
else:
    # We load into globals whatever we have in extension_module
    # We specifically exclude any modules that may be included (like sys, os, etc)
    # *except* for ones that are part of metaflow_custom (basically providing
    # an aliasing mechanism)
    lazy_load_custom_modules = {}
    addl_modules = _ext_plugins.__dict__.get('__mf_promote_submodules__')
    if addl_modules:
        # We make an alias for these modules which the metaflow_custom author
        # wants to expose but that may not be loaded yet
        lazy_load_custom_modules = {
            'metaflow.plugins.%s' % k: 'metaflow_custom.plugins.%s' % k
            for k in addl_modules}
    for n, o in _ext_plugins.__dict__.items():
        if not n.startswith('__') and not isinstance(o, types.ModuleType):
            globals()[n] = o
        elif isinstance(o, types.ModuleType) and o.__package__ and \
                o.__package__.startswith('metaflow_custom'):
            lazy_load_custom_modules['metaflow.plugins.%s' % n] = o
    if lazy_load_custom_modules:
        # NOTE: We load things first to have metaflow_custom override things here.
        # This does mean that for modules that have the same name (for example,
        # if metaflow_custom.plugins also provides a conda module), it needs
        # to provide whatever is expected below (so for example a `conda_step_decorator`
        # file with a `CondaStepDecorator` class).
        # We do this because we want metaflow_custom to fully override things
        # and if we did not change sys.meta_path here, the lines below would
        # load the non metaflow_custom modules providing for possible confusion.
        # This keeps it cleaner.
        from metaflow import _LazyLoader
        sys.meta_path = [_LazyLoader(lazy_load_custom_modules)] + sys.meta_path
    
    class _wrap(object):
        def __init__(self, obj):
            self.__dict__ = obj.__dict__

        def __getattr__(self, name):
            if name in _expected_extensions:
                return _expected_extensions[name]
            raise AttributeError

    _ext_plugins = _wrap(_ext_plugins)



def get_plugin_cli():
    # it is important that CLIs are not imported when
    # __init__ is imported. CLIs may use e.g.
    # parameters.add_custom_parameters which requires
    # that the flow is imported first

    # Add new CLI commands in this list
    from . import package_cli
    from .aws.batch import batch_cli
    from .aws.step_functions import step_functions_cli

    return _ext_plugins.get_plugin_cli() + [
        package_cli.cli,
        batch_cli.cli,
        step_functions_cli.cli]


def _merge_lists(base, overrides, attr):
    # Merge two lists of classes by comparing them for equality using 'attr'.
    # This function prefers anything in 'overrides'. In other words, if a class
    # is present in overrides and matches (according to the equality criterion) a class in
    # base, it will be used instead of the one in base.
    l = list(overrides)
    existing = set([getattr(o, attr) for o in overrides])
    l.extend([d for d in base if getattr(d, attr) not in existing])
    return l

# Add new decorators in this list
from .catch_decorator import CatchDecorator
from .timeout_decorator import TimeoutDecorator
from .environment_decorator import EnvironmentDecorator
from .retry_decorator import RetryDecorator
from .resources_decorator import ResourcesDecorator
from .aws.batch.batch_decorator import BatchDecorator
from .aws.step_functions.step_functions_decorator \
                import StepFunctionsInternalDecorator
from .test_unbounded_foreach_decorator\
    import InternalTestUnboundedForeachDecorator, InternalTestUnboundedForeachInput
from .conda.conda_step_decorator import CondaStepDecorator

STEP_DECORATORS = _merge_lists([CatchDecorator,
                                TimeoutDecorator,
                                EnvironmentDecorator,
                                ResourcesDecorator,
                                RetryDecorator,
                                BatchDecorator,
                                StepFunctionsInternalDecorator,
                                CondaStepDecorator,
                                InternalTestUnboundedForeachDecorator],
                                    _ext_plugins.STEP_DECORATORS, 'name')

# Add Conda environment
from .conda.conda_environment import CondaEnvironment
ENVIRONMENTS = _merge_lists([CondaEnvironment], _ext_plugins.ENVIRONMENTS, 'TYPE')

# Metadata providers
from .metadata import LocalMetadataProvider, ServiceMetadataProvider

METADATA_PROVIDERS = _merge_lists(
    [LocalMetadataProvider, ServiceMetadataProvider], _ext_plugins.METADATA_PROVIDERS, 'TYPE')

# Every entry in this list becomes a class-level flow decorator.
# Add an entry here if you need a new flow-level annotation. Be
# careful with the choice of name though - they become top-level
# imports from the metaflow package.
from .conda.conda_flow_decorator import CondaFlowDecorator
from .aws.step_functions.schedule_decorator import ScheduleDecorator
from .project_decorator import ProjectDecorator
FLOW_DECORATORS = _merge_lists([CondaFlowDecorator,
                                ScheduleDecorator,
                                ProjectDecorator],
                            _ext_plugins.FLOW_DECORATORS, 'name')


# Sidecars
from ..mflog.save_logs_periodically import SaveLogsPeriodicallySidecar
from metaflow.metadata.heartbeat import MetadataHeartBeat

SIDECARS = {'save_logs_periodically': SaveLogsPeriodicallySidecar,
            'heartbeat': MetadataHeartBeat}
SIDECARS.update(_ext_plugins.SIDECARS)

# Add logger
from .debug_logger import DebugEventLogger
LOGGING_SIDECARS = {'debugLogger': DebugEventLogger,
                    'nullSidecarLogger': None}
LOGGING_SIDECARS.update(_ext_plugins.LOGGING_SIDECARS)

# Add monitor
from .debug_monitor import DebugMonitor
MONITOR_SIDECARS = {'debugMonitor': DebugMonitor,
                    'nullSidecarMonitor': None}
MONITOR_SIDECARS.update(_ext_plugins.MONITOR_SIDECARS)

SIDECARS.update(LOGGING_SIDECARS)
SIDECARS.update(MONITOR_SIDECARS)

from .aws.aws_client import Boto3ClientProvider
AWS_CLIENT_PROVIDERS = _merge_lists(
    [Boto3ClientProvider], _ext_plugins.AWS_CLIENT_PROVIDERS, 'name')

# Erase all temporary names to avoid leaking things
# We leave '_ext_plugins' and '_expected_extensions' because they are used in
# a function (so they need to stick around)
for _n in ['ver', 'n', 'o', 'e', 'lazy_load_custom_modules', '_LazyLoader',
           '_merge_lists', '_fake', '_wrap', 'addl_modules']:
    try:
        del globals()[_n]
    except KeyError:
        pass
del globals()['_n']



import os
import sys
from hashlib import sha1
from multiprocessing.dummy import Pool
import platform
import requests
import shutil
import tempfile
try:
    from urlparse import urlparse
except:
    from urllib.parse import urlparse


from metaflow.datastore.local import LocalDataStore
from metaflow.decorators import StepDecorator
from metaflow.metaflow_environment import InvalidEnvironmentException
from metaflow.metadata import MetaDatum
from metaflow.metaflow_config import get_pinned_conda_libs, CONDA_PACKAGE_S3ROOT
from metaflow.util import get_metaflow_root
from metaflow.datatools import S3
from metaflow.unbounded_foreach import UBF_CONTROL

from ..env_escape import generate_trampolines
from . import read_conda_manifest, write_to_conda_manifest
from .conda import Conda

try:
    unicode
except NameError:
    unicode = str
    basestring = str

class CondaStepDecorator(StepDecorator):
    """
    Conda decorator that sets the Conda environment for your step

    To use, add this decorator to your step:
    ```
    @conda
    @step
    def MyStep(self):
        ...
    ```

    Information in this decorator will override any eventual @conda_base flow level decorator.
    Parameters
    ----------
    libraries : Dict
        Libraries to use for this flow. The key is the name of the package and the value
        is the version to use. Defaults to {}
    python : string
        Version of Python to use (for example: '3.7.4'). Defaults to None
        (will use the current python version)
    disabled : bool
        If set to True, disables Conda. Defaults to False
    """
    name = 'conda'
    defaults = {'libraries': {},
                'python': None,
                'disabled': None}

    conda = None
    environments = None

    def _get_base_attributes(self):
        if 'conda_base' in self.flow._flow_decorators:
            return self.flow._flow_decorators['conda_base'].attributes
        return self.defaults

    def _python_version(self):
        return next(x for x in [
                    self.attributes['python'], 
                    self.base_attributes['python'], 
                    platform.python_version()] if x is not None)

    def is_enabled(self, ubf_context=None):
        if ubf_context == UBF_CONTROL:
            # Disable `@conda` for ubf_control tasks.
            return False
        return not next(x for x in [
                        self.attributes['disabled'],
                        self.base_attributes['disabled'],
                        False] if x is not None)

    def _lib_deps(self):
        deps = get_pinned_conda_libs(self._python_version())

        base_deps = self.base_attributes['libraries']
        deps.update(base_deps)
        step_deps = self.attributes['libraries']
        if isinstance(step_deps, (unicode, basestring)):
            step_deps = step_deps.strip('"{}\'')
            if step_deps:
                step_deps = dict(map(lambda x: x.strip().strip('"\''),
                                    a.split(':')) for a in step_deps.split(','))
        deps.update(step_deps)
        return deps

    def _step_deps(self):
        deps = [b'python==%s' % self._python_version().encode()]
        deps.extend(b'%s==%s' % (name.encode('ascii'), ver.encode('ascii'))
                    for name, ver in self._lib_deps().items())
        return deps

    def _env_id(self):
        deps = self._step_deps()
        return 'metaflow_%s_%s_%s' % (self.flow.name, 
                                        self.architecture, 
                                        sha1(b' '.join(sorted(deps))).hexdigest())

    def _resolve_step_environment(self, ds_root, force=False):
        env_id = self._env_id()
        cached_deps = read_conda_manifest(ds_root, self.flow.name)
        if CondaStepDecorator.conda is None:
            CondaStepDecorator.conda = Conda()
            CondaStepDecorator.environments =\
                CondaStepDecorator.conda.environments(self.flow.name)
        if force or env_id not in cached_deps or 'cache_urls' not in cached_deps[env_id]:
            if force or env_id not in cached_deps:
                deps = self._step_deps()
                (exact_deps, urls, order) = \
                    self.conda.create(self.step,
                                      env_id,
                                      deps,
                                      architecture=self.architecture,
                                      disable_safety_checks=self.disable_safety_checks)
                payload = {
                    'explicit': exact_deps,
                    'deps': [d.decode('ascii') for d in deps],
                    'urls': urls,
                    'order': order
                }
            else:
                payload = cached_deps[env_id]
            if self.datastore.TYPE == 's3' and 'cache_urls' not in payload:
                payload['cache_urls'] = self._cache_env()
            write_to_conda_manifest(ds_root, self.flow.name, env_id, payload)
            CondaStepDecorator.environments =\
                CondaStepDecorator.conda.environments(self.flow.name)
        return env_id

    def _cache_env(self):
        def _download(entry):
            url, local_path = entry
            with requests.get(url, stream=True) as r:
                with open(local_path, 'wb') as f:
                    shutil.copyfileobj(r.raw, f)

        env_id = self._env_id()
        files = []
        to_download = []
        for package_info in self.conda.package_info(env_id):
            url = urlparse(package_info['url'])
            path = os.path.join(CONDA_PACKAGE_S3ROOT, 
                                url.netloc,
                                url.path.lstrip('/'),
                                package_info['md5'],
                                package_info['fn'])
            tarball_path = package_info['package_tarball_full_path']
            if tarball_path.endswith('.conda'):
                # Conda doesn't set the metadata correctly for certain fields
                # when the underlying OS is spoofed.
                tarball_path = tarball_path[:-6]
            if not tarball_path.endswith('.tar.bz2'):
                tarball_path = '%s.tar.bz2' % tarball_path
            if not os.path.isfile(tarball_path):
                # The tarball maybe missing when user invokes `conda clean`!
                to_download.append((package_info['url'], tarball_path))
            files.append((path, tarball_path))
        if to_download:
            Pool(8).map(_download, to_download)
        with S3() as s3:
            s3.put_files(files, overwrite=False)
        return [files[0] for files in files]

    def _prepare_step_environment(self, step_name, ds_root):
        env_id = self._resolve_step_environment(ds_root)
        if env_id not in CondaStepDecorator.environments:
            cached_deps = read_conda_manifest(ds_root, self.flow.name)
            self.conda.create(self.step,
                              env_id,
                              cached_deps[env_id]['urls'],
                              architecture=self.architecture,
                              explicit=True,
                              disable_safety_checks=self.disable_safety_checks)
            CondaStepDecorator.environments =\
                CondaStepDecorator.conda.environments(self.flow.name)
        return env_id

    def _disable_safety_checks(self, decos):
        # Disable conda safety checks when creating linux-64 environments on
        # a macOS. This is needed because of gotchas around inconsistently 
        # case-(in)sensitive filesystems for macOS and linux.
        for deco in decos:
            if deco.name == 'batch' and platform.system() == 'Darwin':
                return True
        return False

    def _architecture(self, decos):
        for deco in decos:
            if deco.name == 'batch':
                # force conda resolution for linux-64 architectures
                return 'linux-64'
        bit = '32'
        if platform.machine().endswith('64'):
            bit = '64'
        if platform.system() == 'Linux':
            return 'linux-%s' % bit
        elif platform.system() == 'Darwin':
            return 'osx-%s' % bit
        else:
            raise InvalidEnvironmentException('The *@conda* decorator is not supported '
                                              'outside of Linux and Darwin platforms')

    def runtime_init(self, flow, graph, package, run_id):
        # Create a symlink to installed version of metaflow to execute user code against
        path_to_metaflow = os.path.join(get_metaflow_root(), 'metaflow')
        self.metaflow_home = tempfile.mkdtemp(dir='/tmp')
        self.addl_paths = None
        os.symlink(path_to_metaflow, os.path.join(self.metaflow_home, 'metaflow'))
        # Do the same for metaflow_custom
        try:
            import metaflow_custom as m
        except ImportError:
            # No additional check needed because if we are here, we already checked
            # for other issues when loading at the toplevel
            pass
        else:
            custom_paths = list(m.__path__)
            if len(custom_paths) == 1:
                # Regular package
                os.symlink(custom_paths[0], os.path.join(self.metaflow_home, 'metaflow_custom'))
            else:
                # Namespace package; we don't symlink but add the additional paths
                # for the conda interpreter
                self.addl_paths = [os.path.split(p)[0] for p in custom_paths]

        # Also install any environment escape overrides directly here to enable
        # the escape to work even in non metaflow-created subprocesses
        generate_trampolines(self.metaflow_home)


    def step_init(self, flow, graph, step, decos, environment, datastore, logger):
        if environment.TYPE != 'conda':
            raise InvalidEnvironmentException('The *@conda* decorator requires '
                                              '--environment=conda')
        def _logger(line, **kwargs):
            logger(line)
        self.local_root = LocalDataStore.get_datastore_root_from_config(_logger)
        environment.set_local_root(self.local_root)
        self.architecture = self._architecture(decos)
        self.disable_safety_checks = self._disable_safety_checks(decos)
        self.step = step
        self.flow = flow
        self.datastore = datastore
        self.base_attributes = self._get_base_attributes()
        os.environ['PYTHONNOUSERSITE'] = '1'

    def package_init(self, flow, step, environment):
        if self.is_enabled():
            self._prepare_step_environment(step, self.local_root)

    def runtime_task_created(self,
                             datastore,
                             task_id,
                             split_index,
                             input_paths,
                             is_cloned,
                             ubf_context):
        if self.is_enabled(ubf_context):
            self.env_id = \
                self._prepare_step_environment(self.step, self.local_root)

    def task_pre_step(self,
                      step_name,
                      ds,
                      meta,
                      run_id,
                      task_id,
                      flow,
                      graph,
                      retry_count,
                      max_retries,
                      ubf_context,
                      inputs):
        if self.is_enabled(ubf_context):
            meta.register_metadata(run_id, step_name, task_id,
                                       [MetaDatum(field='conda_env_id',
                                                  value=self._env_id(),
                                                  type='conda_env_id',
                                                  tags=[])])

    def runtime_step_cli(self,
                         cli_args,
                         retry_count,
                         max_user_code_retries,
                         ubf_context):
        if self.is_enabled(ubf_context) and 'batch' not in cli_args.commands:
            python_path = self.metaflow_home
            if os.environ.get('PYTHONPATH') is not None:
                python_path = os.pathsep.join([os.environ['PYTHONPATH'], python_path])
            if self.addl_paths is not None:
                addl_paths = os.pathsep.join(self.addl_paths)
                python_path = os.pathsep.join([addl_paths, python_path])
            env_path = os.path.dirname(self.conda.python(self.env_id))
            if os.environ.get('PATH') is not None:
                env_path = os.pathsep.join([env_path, os.environ['PATH']])
            
            cli_args.env['PATH'] = env_path
            cli_args.env['PYTHONPATH'] = python_path
            cli_args.env['_METAFLOW_CONDA_ENV'] = self.env_id
            cli_args.entrypoint[0] = self.conda.python(self.env_id)

    def runtime_finished(self, exception):
        shutil.rmtree(self.metaflow_home)



import os

# This script is used to upload logs during task bootstrapping, so
# it shouldn't have external dependencies besides Metaflow itself
# (e.g. no click for parsing CLI args).
from metaflow.datastore import DATASTORES
from metaflow.util import Path
from . import TASK_LOG_SOURCE

SMALL_FILE_LIMIT = 1024*1024

def save_logs():
    def _read_file(path):
        with open(path, 'rb') as f:
            return f.read()

    # these env vars are set by mflog.mflog_env
    pathspec = os.environ['MF_PATHSPEC']
    attempt = os.environ['MF_ATTEMPT']
    ds_type = os.environ['MF_DATASTORE']
    ds_root = os.environ.get('MF_DATASTORE_ROOT')
    paths = (os.environ['MFLOG_STDOUT'],\
             os.environ['MFLOG_STDERR'])

    flow_name, run_id, step_name, task_id = pathspec.split('/')
    Datastore = DATASTORES[ds_type]
    if ds_root is None:
        def print_clean(line, **kwargs):
            pass
        ds_root = Datastore.get_datastore_root_from_config(print_clean)
    Datastore.datastore_root = ds_root

    ds = Datastore(flow_name,
                   run_id=run_id,
                   step_name=step_name,
                   task_id=task_id,
                   attempt=int(attempt),
                   mode='w')

    try:
        streams = ('stdout', 'stderr')
        sizes = [(stream, path, os.path.getsize(path))
                 for stream, path in zip(streams, paths)
                 if os.path.exists(path)]

        if max(size for _, _, size in sizes) < SMALL_FILE_LIMIT:
            op = _read_file
        else:
            op = Path

        data = [(stream, op(path)) for stream, path, _ in sizes]
        ds.save_logs(TASK_LOG_SOURCE, data)
    except:
        # Upload failing is not considered a fatal error.
        # This script shouldn't return non-zero exit codes
        # for transient errors.
        pass

if __name__ == '__main__':
    save_logs()
    #to debug delays in logs, comment the line above and uncomment
    #this snippet:
    """
    import sys
    from metaflow.metaflow_profile import profile
    d = {}
    with profile('save_logs', stats_dict=d):
        save_logs()
    print('Save logs took %dms' % d['save_logs'], file=sys.stderr)
    """